"""
Functions for network matched-filter detection of seismic data.

Designed to cross-correlate templates generated by template_gen function
with data and output the detections.

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import logging

import numpy as np
from obspy import Stream

from eqcorrscan.core.match_filter.helpers import extract_from_stream

Logger = logging.getLogger(__name__)


class MatchFilterError(Exception):
    """
    Default error for match-filter errors.
    """

    def __init__(self, value):
        """
        Raise error.

        .. rubric:: Example

        >>> MatchFilterError('This raises an error')
        This raises an error
        """
        self.value = value

    def __repr__(self):
        """
        Print the value of the error.

        .. rubric:: Example
        >>> print(MatchFilterError('Error').__repr__())
        Error
        """
        return self.value

    def __str__(self):
        """
        Print the error in a pretty way.

        .. rubric:: Example
        >>> print(MatchFilterError('Error'))
        Error
        """
        return self.value


# Note: maintained for backwards compatability. All efforts now in tribes
def match_filter(template_names, template_list, st, threshold,
                 threshold_type, trig_int, plot=False, plotdir=None,
                 xcorr_func=None, concurrency=None, cores=None,
                 plot_format='png', output_cat=False,
                 extract_detections=False, arg_check=True, full_peaks=False,
                 peak_cores=None, export_cccsums=False, **kwargs):
    """
    Main matched-filter detection function.

    Over-arching code to run the correlations of given templates with a
    day of seismic data and output the detections based on a given threshold.
    For a functional example see the tutorials.

    :type template_names: list
    :param template_names:
        List of template names in the same order as template_list
    :type template_list: list
    :param template_list:
        A list of templates of which each template is a
        :class:`obspy.core.stream.Stream` of obspy traces containing seismic
        data and header information.
    :type st: :class:`obspy.core.stream.Stream`
    :param st:
        A Stream object containing all the data available and
        required for the correlations with templates given.  For efficiency
        this should contain no excess traces which are not in one or more of
        the templates.  This will now remove excess traces internally, but
        will copy the stream and work on the copy, leaving your input stream
        untouched.
    :type threshold: float
    :param threshold: A threshold value set based on the threshold_type
    :type threshold_type: str
    :param threshold_type:
        The type of threshold to be used, can be MAD, absolute or av_chan_corr.
        See Note on thresholding below.
    :type trig_int: float
    :param trig_int:
        Minimum gap between detections from one template in seconds.
        If multiple detections occur within trig_int of one-another, the one
        with the highest cross-correlation sum will be selected.
    :type plot: bool
    :param plot: Turn plotting on or off
    :type plotdir: str
    :param plotdir:
        Path to plotting folder, plots will be output here, defaults to None,
        and plots are shown on screen.
    :type xcorr_func: str or callable
    :param xcorr_func:
        A str of a registered xcorr function or a callable for implementing
        a custom xcorr function. For more information see:
        :func:`eqcorrscan.utils.correlate.register_array_xcorr`
    :type concurrency: str
    :param concurrency:
        The type of concurrency to apply to the xcorr function. Options are
        'multithread', 'multiprocess', 'concurrent'. For more details see
        :func:`eqcorrscan.utils.correlate.get_stream_xcorr`
    :type cores: int
    :param cores: Number of cores to use
    :type plot_format: str
    :param plot_format: Specify format of output plots if saved
    :type output_cat: bool
    :param output_cat:
        Specifies if matched_filter will output an obspy.Catalog class
        containing events for each detection. Default is False, in which case
        matched_filter will output a list of detection classes, as normal.
    :type output_event: bool
    :param output_event:
        Whether to include events in the Detection objects, defaults to True,
        but for large cases you may want to turn this off as Event objects
        can be quite memory intensive.
    :type extract_detections: bool
    :param extract_detections:
        Specifies whether or not to return a list of streams, one stream per
        detection.
    :type arg_check: bool
    :param arg_check:
        Check arguments, defaults to True, but if running in bulk, and you are
        certain of your arguments, then set to False.
    :type full_peaks: bool
    :param full_peaks: See
        :func: `eqcorrscan.utils.findpeaks.find_peaks_compiled`
    :type peak_cores: int
    :param peak_cores:
        Number of processes to use for parallel peak-finding (if different to
        `cores`).
    :type spike_test: bool
    :param spike_test: If set True, raise error when there is a spike in data.
        defaults to True.
    :type copy_data: bool
    :param copy_data:
        Whether to copy data to keep it safe, otherwise will edit your
        templates and stream in place.
    :type export_cccsums: bool
    :param export_cccsums: Whether to save the cross-correlation statistic.


    .. Note::
        When using the "fftw" correlation backend the length of the fft
        can be set. See :mod:`eqcorrscan.utils.correlate` for more info.

    .. note::
        **Returns:**

        If neither `output_cat` or `extract_detections` are set to `True`,
        then only the list of :class:`eqcorrscan.core.match_filter.Detection`'s
        will be output:

        :return:
            :class:`eqcorrscan.core.match_filter.Detection` detections for each
            detection made.
        :rtype: list

        If `output_cat` is set to `True`, then the
        :class:`obspy.core.event.Catalog` will also be output:

        :return: Catalog containing events for each detection, see above.
        :rtype: :class:`obspy.core.event.Catalog`

        If `extract_detections` is set to `True` then the list of
        :class:`obspy.core.stream.Stream`'s will also be output.

        :return:
            list of :class:`obspy.core.stream.Stream`'s for each detection, see
            above.
        :rtype: list

    .. note::
        If your data contain gaps these must be padded with zeros before
        using this function. The `eqcorrscan.utils.pre_processing` functions
        will provide gap-filled data in the appropriate format.  Note that if
        you pad your data with zeros before filtering or resampling the gaps
        will not be all zeros after filtering. This will result in the
        calculation of spurious correlations in the gaps.

    .. Note::
        Detections are not corrected for `pre-pick`, the
        detection.detect_time corresponds to the beginning of the earliest
        template channel at detection.

    .. note::
        **Data overlap:**

        Internally this routine shifts and trims the data according to the
        offsets in the template (e.g. if trace 2 starts 2 seconds after trace 1
        in the template then the continuous data will be shifted by 2 seconds
        to align peak correlations prior to summing).  Because of this,
        detections at the start and end of continuous data streams
        **may be missed**.  The maximum time-period that might be missing
        detections is the maximum offset in the template.

        To work around this, if you are conducting matched-filter detections
        through long-duration continuous data, we suggest using some overlap
        (a few seconds, on the order of the maximum offset in the templates)
        in the continous data.  You will then need to post-process the
        detections (which should be done anyway to remove duplicates).

    .. note::
        **Thresholding:**

        **MAD** threshold is calculated as the:

        .. math::

            threshold {\\times} (median(abs(cccsum)))

        where :math:`cccsum` is the cross-correlation sum for a given template.

        **absolute** threshold is a true absolute threshold based on the
        cccsum value.

        **av_chan_corr** is based on the mean values of single-channel
        cross-correlations assuming all data are present as required for the
        template, e.g:

        .. math::

            av_chan_corr_thresh=threshold \\times (cccsum/len(template))

        where :math:`template` is a single template from the input and the
        length is the number of channels within this template.

    .. note::
        The output_cat flag will create an :class:`obspy.core.event.Catalog`
        containing one event for each
        :class:`eqcorrscan.core.match_filter.Detection`'s generated by
        match_filter. Each event will contain a number of comments dealing
        with correlation values and channels used for the detection. Each
        channel used for the detection will have a corresponding
        :class:`obspy.core.event.Pick` which will contain time and
        waveform information. **HOWEVER**, the user should note that
        the pick times do not account for the prepick times inherent in
        each template. For example, if a template trace starts 0.1 seconds
        before the actual arrival of that phase, then the pick time generated
        by match_filter for that phase will be 0.1 seconds early.
    """
    from eqcorrscan.core.match_filter import Tribe, Template

    if "plotvar" in kwargs.keys():
        Logger.warning("plotvar is depreciated, use plot instead")
        plot = kwargs.get("plotvar")

    if arg_check:
        # Check the arguments to be nice - if arguments wrong type the parallel
        # output for the error won't be useful
        if not isinstance(template_names, list):
            raise MatchFilterError('template_names must be of type: list')
        if not isinstance(template_list, list):
            raise MatchFilterError('templates must be of type: list')
        if not len(template_list) == len(template_names):
            raise MatchFilterError('Not the same number of templates as names')
        for template in template_list:
            if not isinstance(template, Stream):
                msg = 'template in template_list must be of type: ' + \
                      'obspy.core.stream.Stream'
                raise MatchFilterError(msg)
        if not isinstance(st, Stream):
            msg = 'st must be of type: obspy.core.stream.Stream'
            raise MatchFilterError(msg)
        if str(threshold_type) not in [str('MAD'), str('absolute'),
                                       str('av_chan_corr')]:
            msg = 'threshold_type must be one of: MAD, absolute, av_chan_corr'
            raise MatchFilterError(msg)
        for tr in st:
            if not tr.stats.sampling_rate == st[0].stats.sampling_rate:
                raise MatchFilterError('Sampling rates are not equal %f: %f' %
                                       (tr.stats.sampling_rate,
                                        st[0].stats.sampling_rate))
        for template in template_list:
            for tr in template:
                if not tr.stats.sampling_rate == st[0].stats.sampling_rate:
                    raise MatchFilterError(
                        'Template sampling rate does not '
                        'match continuous data')
        for template in template_list:
            for tr in template:
                if isinstance(tr.data, np.ma.core.MaskedArray):
                    raise MatchFilterError(
                        'Template contains masked array, split first')

    # Make a tribe and run tribe.detect
    tribe = Tribe()
    # Cope with naming issues
    name_mapper = {template_name: f"template_{i}"
                   for i, template_name in enumerate(template_names)}
    for template, template_name in zip(template_list, template_names):
        tribe += Template(
            st=template, name=name_mapper[template_name],
            process_length=(st[0].stats.endtime - st[0].stats.starttime),
            prepick=0.0, samp_rate=template[0].stats.sampling_rate,
        )

    # Data must be pre-processed
    party = tribe.detect(
        stream=st, threshold=threshold, threshold_type=threshold_type,
        trig_int=trig_int, plot=plot, plotdir=plotdir, daylong=False,
        parallel_process=False, xcorr_func=xcorr_func, concurrency=concurrency,
        cores=cores, ignore_length=True, ignore_bad_data=True, group_size=None,
        overlap="calculate", full_peaks=full_peaks, save_progress=False,
        process_cores=None, pre_processed=True, check_processing=False,
        return_stream=False, plot_format=plot_format,
        peak_cores=peak_cores, export_cccsums=export_cccsums, **kwargs
    )
    detections = [d for f in party for d in f]

    # Remap template names
    name_mapper = {val: key for key, val in name_mapper.items()}
    for d in detections:
        d.template_name = name_mapper[d.template_name]

    Logger.info("Made {0} detections from {1} templates".format(
        len(detections), len(tribe)))
    if extract_detections:
        detection_streams = extract_from_stream(st, detections)

    if output_cat and not extract_detections:
        return detections, party.get_catalog()
    elif not extract_detections:
        return detections
    elif extract_detections and not output_cat:
        return detections, detection_streams
    else:
        return detections, party.get_catalog(), detection_streams


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Test the helpers for EQcorrscan's matched-filter codes.
"""

import abc
import copy
import os.path
import time
import shutil
import tempfile
import logging
import unittest
import pytest

from multiprocessing import Process, Queue

from obspy import UTCDateTime, read
from obspy.clients.fdsn import Client
from obspy.clients.earthworm import Client as EWClient

from eqcorrscan.core.match_filter import Party
from eqcorrscan.core.match_filter.helpers import (
    get_waveform_client, _test_event_similarity)
from eqcorrscan.core.match_filter.helpers.processes import (
    _get_detection_stream, Poison, _prepper, _pre_processor,
    _make_detections, _FDSN_parseable_Client)

try:
    from pytest_cov.embed import cleanup_on_sigterm
except ImportError:
    pass
else:
    cleanup_on_sigterm()


Logger = logging.getLogger(__name__)
MAX_WAIT = 15  # Maximum wait time for process to close


def get_test_templates():
    party = Party().read(
        filename=os.path.join(
            os.path.abspath(os.path.dirname(os.path.dirname(__file__))),
            'test_data', 'test_party.tgz'))
    return [f.template for f in party]


class TestHelperFunctions(unittest.TestCase):
    def test_monkey_patching(self):
        """ Test that monkey patching a client works. """
        client = EWClient("pubavo1.wr.usgs.gov", 16022)
        self.assertFalse(hasattr(client, "get_waveforms_bulk"))
        client = get_waveform_client(client)
        self.assertTrue(hasattr(client, "get_waveforms_bulk"))

    @pytest.mark.network
    def test_event_similarity_quiet(self):
        self._event_similarity(verbose=False)

    @pytest.mark.network
    def test_event_similarity_loud(self):
        self._event_similarity(verbose=True)

    def _event_similarity(self, verbose: bool = False):
        client = Client("GEONET")
        event = client.get_events(eventid="2023p923930")[0]
        self.assertTrue(_test_event_similarity(
            event, event, verbose=verbose))
        event2 = event.copy()
        self.assertTrue(_test_event_similarity(
            event, event2, verbose=verbose))
        with self.assertRaises(NotImplementedError):
            _test_event_similarity(event, "bob")
        event2.origins = event2.origins[0:-2]
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        event2 = event.copy()
        event2.origins[-1].arrivals = event2.origins[-1].arrivals[0:-2]
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        event2 = event.copy()
        event2.origins[-1].arrivals[-1].time_residual = \
            event2.origins[-1].arrivals[-1].time_residual - .5
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        event2 = event.copy()
        event2.origins[-1].arrivals[-1].distance = \
            event2.origins[-1].arrivals[-1].distance - 1.5
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        event2 = event.copy()
        event2.origins[-1].time -= 60
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        # Picks
        event2 = event.copy()
        event2.picks = event2.picks[0:-2]
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        event2 = event.copy()
        event2.picks[0].time += 20
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        event2 = event.copy()
        event2.picks[0].waveform_id.station_code = "BOB"
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        event2 = event.copy()
        event2.picks[0].waveform_id.channel_code = "BOB"
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        # Amplitudes
        event2 = event.copy()
        event2.amplitudes = event2.amplitudes[0:-2]
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        event2 = event.copy()
        event2.amplitudes[0].generic_amplitude += 100
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        event2 = event.copy()
        event2.amplitudes[0].waveform_id.station_code = "BOB"
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))
        event2 = event.copy()
        event2.amplitudes[0].waveform_id.channel_code = "BOB"
        self.assertFalse(_test_event_similarity(
            event, event2, verbose=verbose))


class ProcessTests(abc.ABC, unittest.TestCase):
    directories_to_nuke = []
    wait_time = 5
    kwargs = dict()
    process = None

    @classmethod
    def tearDownClass(cls):
        for _dir in cls.directories_to_nuke:
            try:
                shutil.rmtree(_dir)
            except FileNotFoundError:
                pass

    def wait_for_close(self):
        total_wait = 0
        while total_wait <= MAX_WAIT:
            if self.process.is_alive():
                time.sleep(self.wait_time)
            else:
                break
        self.assertFalse(self.process.is_alive())
        self.process.join()


class TestMakeDetections(ProcessTests):
    @classmethod
    def setUpClass(cls):
        templates = get_test_templates()
        cls.global_kwargs = dict(
            delta=0.01,
            templates=templates,
            threshold=8.0,
            threshold_type="MAD",
            save_progress=False,
        )

    def setUp(self):
        self.kwargs = copy.copy(self.global_kwargs)
        self.kwargs.update(dict(
            input_queue=Queue(),
            output_queue=Queue(),
            poison_queue=Queue()
        ))

    def test_poisoning(self):
        self.process = Process(
            target=_make_detections, kwargs=self.kwargs,
            name="TestProcess")
        poisoning(obj=self)

    def test_poisoning_from_input(self):
        self.process = Process(
            target=_make_detections, kwargs=self.kwargs,
            name="TestProcess")
        poisoning_from_input(
            obj=self, input_queue=self.kwargs['input_queue'])


class TestPrepper(ProcessTests):
    @classmethod
    def setUpClass(cls):
        templates = get_test_templates()
        cls.global_kwargs = dict(
            templates=templates,
            group_size=5,
            groups=None,
            xcorr_func="fmf",
        )

    def setUp(self):
        self.kwargs = copy.copy(self.global_kwargs)
        self.kwargs.update(dict(
            input_stream_filename_queue=Queue(),
            output_queue=Queue(maxsize=1),
            poison_queue=Queue()
        ))

    def test_poisoning(self):
        self.process = Process(
            target=_prepper, kwargs=self.kwargs,
            name="TestProcess")
        poisoning(obj=self)

    def test_poisoning_while_waiting_on_output(self):
        self.process = Process(
            target=_prepper, kwargs=self.kwargs,
            name="TestProcess")
        poisoning_while_waiting_on_output(
            obj=self, output_queue=self.kwargs["output_queue"])

    def test_poisoning_from_input(self):
        self.process = Process(
            target=_prepper, kwargs=self.kwargs,
            name="TestProcess")
        poisoning_from_input(
            obj=self, input_queue=self.kwargs['input_stream_filename_queue'])


class TestPreProcessor(ProcessTests):
    @classmethod
    def setUpClass(cls):
        process_length = 360
        cls.global_kwargs = dict(
            template_ids={("NZ", "WVZ", "10", "HHZ")},
            pre_processed=False,
            ignore_length=False,
            ignore_bad_data=False,
            filt_order=4,
            highcut=20,
            lowcut=2,
            samp_rate=50,
            process_length=process_length,
            parallel=False,
            cores=1,
            daylong=False,
            overlap=3.0,
        )

    def setUp(self):
        self.kwargs = copy.copy(self.global_kwargs)
        self.kwargs.update(dict(
            input_stream_queue=Queue(),
            temp_stream_dir=tempfile.mkdtemp(),
            output_filename_queue=Queue(maxsize=1),
            poison_queue=Queue()
        ))
        Logger.info(self.kwargs)
        self.directories_to_nuke.append(self.kwargs['temp_stream_dir'])

    def test_poisoning(self):
        self.process = Process(
            target=_pre_processor, kwargs=self.kwargs,
            name="TestProcess")
        poisoning(obj=self)

    def test_poisoning_while_waiting_on_output(self):
        self.process = Process(
            target=_pre_processor, kwargs=self.kwargs,
            name="TestProcess")
        poisoning_while_waiting_on_output(
            obj=self, output_queue=self.kwargs["output_filename_queue"])

    def test_poisoning_from_input(self):
        self.process = Process(
            target=_pre_processor, kwargs=self.kwargs,
            name="TestProcess")
        poisoning_from_input(
            obj=self, input_queue=self.kwargs['input_stream_queue'])


class TestGetDetectionStreamProcess(ProcessTests):
    @classmethod
    def setUpClass(cls):
        process_length = 360
        cls.global_kwargs = dict(
            template_channel_ids=[("NZ", "WVZ", "10", "HHZ")],
            client=_FDSN_parseable_Client.from_client(Client("GEONET")),
            retries=3,
            min_gap=0.0,
            buff=3,
            full_stream_dir=None,
            pre_process=False,
            parallel_process=True,
            process_cores=1,
            daylong=False,
            overlap=3.0,
            ignore_length=False,
            ignore_bad_data=False,
            filt_order=4,
            highcut=20,
            lowcut=2,
            samp_rate=50,
            process_length=process_length,
        )

    def setUp(self):
        # Make a copy of the class-wide kwargs
        self.kwargs = copy.copy(self.global_kwargs)
        self.kwargs.update(
            dict(input_time_queue=Queue(),
                 poison_queue=Queue(),
                 output_filename_queue=Queue(maxsize=1),
                 temp_stream_dir=tempfile.mkdtemp()))
        Logger.info(self.kwargs)
        # Cleanup
        self.directories_to_nuke.append(self.kwargs['temp_stream_dir'])

    def test_poisoning(self):
        self.process = Process(
            target=_get_detection_stream, kwargs=self.kwargs,
            name="TestProcess")
        poisoning(obj=self)

    def test_poisoning_while_waiting_on_output(self):
        self.process = Process(
            target=_get_detection_stream, kwargs=self.kwargs,
            name="TestProcess")
        poisoning_while_waiting_on_output(
            obj=self, output_queue=self.kwargs['output_filename_queue'])

    def test_poisoning_from_input(self):
        self.process = Process(
            target=_get_detection_stream, kwargs=self.kwargs,
            name="TestProcess")
        poisoning_from_input(
            obj=self, input_queue=self.kwargs['input_time_queue'])

    def test_normal_operation(self):
        self.process = Process(
            target=_get_detection_stream, kwargs=self.kwargs,
            name="TestProcess")
        self.process.start()

        # Populate time queue
        self.kwargs['input_time_queue'].put(
            (UTCDateTime(2021, 1, 1),
             UTCDateTime(2021, 1, 1, 0, 10)))
        self.kwargs['input_time_queue'].put(None)

        # Get the output
        filename = self.kwargs['output_filename_queue'].get()
        self.assertTrue(os.path.isfile(filename))
        self.assertEqual(self.kwargs['output_filename_queue'].get(), None)

        # Wait for the process to end
        self.wait_for_close()

    def test_full_stream_operation(self):
        kwargs = copy.copy(self.kwargs)
        kwargs.update(dict(
            full_stream_dir=tempfile.mkdtemp(),
            pre_process=True,))
        self.process = Process(
            target=_get_detection_stream, kwargs=kwargs,
            name="TestProcess")
        self.process.start()

        # Populate time queue
        kwargs['input_time_queue'].put(
            (UTCDateTime(2021, 1, 1),
             UTCDateTime(2021, 1, 1, 0, 10)))
        kwargs['input_time_queue'].put(None)

        # Get the output
        filename = kwargs['output_filename_queue'].get()
        self.assertTrue(os.path.isfile(filename))
        self.assertEqual(kwargs['output_filename_queue'].get(), None)

        # Wait for the process to end
        self.wait_for_close()

        # Check for full stream
        full_st = read(f"{kwargs['full_stream_dir']}/*")
        st = read(filename)

        self.assertEqual(st[0].id, full_st[0].id)
        self.assertEqual(st[0].stats.sampling_rate, kwargs['samp_rate'])
        self.assertEqual(full_st[0].stats.sampling_rate, 100.0)

        # Cleanup
        self.directories_to_nuke.append(kwargs['full_stream_dir'])

    def test_exception_handling(self):
        kwargs = copy.copy(self.kwargs)
        kwargs.update(dict(
            overlap="bob",  # This need to be a float! Should raise exception
            pre_process=True,
        ))
        self.process = Process(
            target=_get_detection_stream, kwargs=kwargs, name="ProcessProcess")
        self.process.start()

        # Populate time queue
        kwargs['input_time_queue'].put(
            (UTCDateTime(2021, 1, 1),
             UTCDateTime(2021, 1, 1, 0, 10)))
        kwargs['input_time_queue'].put(None)

        time.sleep(self.wait_time)
        poison = kwargs['poison_queue'].get()
        self.assertIsInstance(poison, Poison)
        self.wait_for_close()


###############################################################################
#  STANDARD PROCESS DEATH TESTS
###############################################################################


def poisoning(obj: ProcessTests):
    obj.process.start()
    # Test death
    obj.kwargs['poison_queue'].put(Exception("TestException"))
    obj.wait_for_close()


def poisoning_while_waiting_on_output(obj: ProcessTests, output_queue: Queue):
    # Fill output queue
    output_queue.put("This is dog")
    obj.process.start()
    # Test death
    obj.kwargs['poison_queue'].put(Exception("TestException"))
    obj.wait_for_close()


def poisoning_from_input(obj: ProcessTests, input_queue: Queue):
    obj.process.start()
    # Test death
    input_queue.put(Poison(Exception("TestException")))
    obj.wait_for_close()


if __name__ == '__main__':
    """
    Run core tests
    """
    unittest.main()



"""
Helpers to get library names. Edited for our use from obspy.core.util.libnames

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""

import os
import ctypes
import logging
from distutils import sysconfig
import importlib.machinery


Logger = logging.getLogger(__name__)


def _load_cdll(name):
    """
    Helper function to load a shared library built during installation
    with ctypes.

    :type name: str
    :param name: Name of the library to load (e.g. 'mseed').
    :rtype: :class:`ctypes.CDLL`
    """
    libdir = os.path.join(os.path.dirname(__file__), 'lib')
    # Try and cope with static FFTW libs
    static_fftw = os.path.join(libdir, 'libfftw3-3.dll')
    static_fftwf = os.path.join(libdir, 'libfftw3f-3.dll')
    try:
        fftw_lib = ctypes.CDLL(str(static_fftw))  # noqa: F841
        fftwf_lib = ctypes.CDLL(str(static_fftwf))  # noqa: F841
    except Exception:
        pass

    # Cope with range of possible extensions for different python versions
    errs = []
    for ext in importlib.machinery.EXTENSION_SUFFIXES:
        libpath = os.path.join(libdir, name + ext)
        try:
            cdll = ctypes.CDLL(str(libpath), mode=ctypes.RTLD_LOCAL)
        except Exception as e:
            msg = 'Could not load shared library "%s".\n\n %s' % (name, str(e))
            errs.append(msg)
            Logger.debug(msg)
        else:
            Logger.debug(f"Loaded library from {libpath}")
            return cdll
    raise ImportError(
        "Could not load shared library {0} due to "
        "errors:\n{1}".format(name, "\n".join(errs)))


if __name__ == '__main__':
    cdll = _load_cdll('libutils')



"""
Helper functions for access to archives, e.g. Party, Tribe

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""

import contextlib
import os
import shutil
import tempfile
import tarfile
import logging


Logger = logging.getLogger(__name__)


@contextlib.contextmanager
def temporary_directory():
    """ make a temporary directory, yeild its name, cleanup on exit """
    dir_name = tempfile.mkdtemp()
    yield dir_name
    if os.path.exists(dir_name):
        shutil.rmtree(dir_name)


def _par_read(dirname, compressed=True):
    """
    Internal write function to read a formatted parameter file.

    :type dirname: str
    :param dirname: Directory to read the parameter file from.
    :type compressed: bool
    :param compressed: Whether the directory is compressed or not.
    """
    from eqcorrscan.core.match_filter.matched_filter import MatchFilterError
    from eqcorrscan.core.match_filter.template import Template

    templates = []
    if compressed:
        arc = tarfile.open(dirname, "r:*")
        members = arc.getmembers()
        _parfile = [member for member in members
                    if member.name.split(os.sep)[-1] ==
                    'template_parameters.csv']
        if len(_parfile) == 0:
            arc.close()
            raise MatchFilterError(
                'No template parameter file in archive')
        parfile = arc.extractfile(_parfile[0])
    else:
        parfile = open(dirname + '/' + 'template_parameters.csv', 'r')
    for line in parfile:
        t_in = Template()
        for key_pair in line.rstrip().split(','):
            if key_pair.split(':')[0].strip() == 'name':
                t_in.__dict__[key_pair.split(':')[0].strip()] = \
                    key_pair.split(':')[-1].strip()
            elif key_pair.split(':')[0].strip() == 'filt_order':
                try:
                    t_in.__dict__[key_pair.split(':')[0].strip()] = \
                        int(key_pair.split(':')[-1])
                except ValueError:
                    pass
            else:
                try:
                    t_in.__dict__[key_pair.split(':')[0].strip()] = \
                        float(key_pair.split(':')[-1])
                except ValueError:
                    pass
        templates.append(t_in)
    parfile.close()
    if compressed:
        arc.close()
    return templates


def _resolved(x):
    return os.path.realpath(os.path.abspath(x))


def _badpath(path, base):
    """
    joinpath will ignore base if path is absolute.
    """
    return not _resolved(os.path.join(base, path)).startswith(base)


def _badlink(info, base):
    """
    Links are interpreted relative to the directory containing the link
    """
    tip = _resolved(os.path.join(base, os.path.dirname(info.name)))
    return _badpath(info.linkname, base=tip)


def _safemembers(members):
    """
    Check members of a tar archive for safety.

    Ensure that they do not contain paths or links outside of where we
    need them - this would only happen if the archive wasn't made by
    eqcorrscan.

    :type members: :class:`tarfile.TarFile`
    :param members: an open tarfile.
    """
    base = _resolved(".")

    for finfo in members:
        assert not _badpath(finfo.name, base), \
            f"{finfo.name} is blocked (illegal path)"
        if finfo.issym() or finfo.islnk():
            assert not _badlink(finfo, base), \
                f"{finfo.name} is blocked: Link to {finfo.linkname}"
        else:
            yield finfo


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
#
# EQcorrscan documentation build configuration file, created by
# sphinx-quickstart on Mon Mar 23 21:20:41 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import shlex
sys.path.insert(0, os.path.abspath('../..'))
import matplotlib
import eqcorrscan

import sphinx_bootstrap_theme
# Use mock to allow for autodoc compilation without needing C based modules
import mock
import glob

READ_THE_DOCS = os.environ.get('READTHEDOCS', None) == 'True'
MOCK_MODULES = ['cv2', 'h5py', 'eqcorrscan.utils.libnames']
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = mock.Mock()

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('../core'))
sys.path.insert(0, os.path.abspath('../utils'))

sys.path = [os.path.dirname(__file__) + os.sep + '_ext'] + sys.path


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.1'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.intersphinx',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    # 'matplotlib.sphinxext.mathmpl',
    # 'matplotlib.sphinxext.only_directives',
    'matplotlib.sphinxext.plot_directive',
    'sphinx.ext.mathjax',
    'sphinx.ext.todo',
    # local extensions
    'sphinx.ext.autosummary',
    'obspydoc',
    'nbsphinx'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'EQcorrscan'
copyright = u'2015-2021: EQcorrscan developers'
author = u'EQcorrscan developers'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '0.4'
# The full version, including alpha/beta/rc tags.
release = eqcorrscan.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# File formats to generate.
plot_formats = [('png', 110), ('hires.png', 200)]
if READ_THE_DOCS:
    plot_formats += [('pdf', 200)]
plot_html_show_formats = True

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'bootstrap'
html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()

html_theme_options = {
    'bootswatch_theme': "sandstone",
}

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = 'EQcorrscan'

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = './EQcorrscan_logo.jpg'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = './EQcorrscan_logo.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'EQcorrscandoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    'papersize': 'a4paper',

    # The font size ('10pt', '11pt' or '12pt').
    'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',

    # Latex figure (float) alignment
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  (master_doc, 'EQcorrscan.tex', u'EQcorrscan Documentation',
   u'Calum John Chamberlain', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = 'EQcorrscan_logo.pdf'

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'eqcorrscan', u'EQcorrscan Documentation',
     [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  (master_doc, 'EQcorrscan', u'EQcorrscan Documentation',
   author, 'EQcorrscan', 'One line description of project.',
   'Miscellaneous'),
]

# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/2.7/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'matplotlib': ('http://matplotlib.org/', None),
    'sqlalchemy': ('http://docs.sqlalchemy.org/en/latest/', None),
    'obspy': ('https://docs.obspy.org/', None),
}

# generate automatically stubs
autosummary_generate = glob.glob("submodules" + os.sep + "*.rst")

# Don't merge __init__ method in auoclass content
autoclass_content = 'class'

# This value is a list of autodoc directive flags that should be automatically
# applied to all autodoc directives. The supported flags are 'members',
# 'undoc-members', 'private-members', 'special-members', 'inherited-members' an
# 'show-inheritance'. Don't set it to anything !
autodoc_default_flags = ['show-inheritance']

# warn about *all* references where the target cannot be found
nitpicky = False

trim_doctest_flags = True



"""
Functions for network matched-filter detection of seismic data.

Designed to cross-correlate templates generated by template_gen function
with data and output the detections.

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import os
import tempfile
import time
import traceback
import logging
import numpy as np

from typing import List, Union, Iterable
from timeit import default_timer

from multiprocessing import Queue
from queue import Empty

from obspy import Stream
from obspy.clients.fdsn import Client

from eqcorrscan.core.match_filter.helpers import (
    _pickle_stream, _unpickle_stream)
from eqcorrscan.core.match_filter.helpers.tribe import (
    _download_st, _pre_process, _group, _detect,
    _read_template_db, _make_party)

from eqcorrscan.utils.correlate import (
    _get_array_dicts, _fmf_stabilisation, _fmf_reshape)
from eqcorrscan.utils.pre_processing import (
    _quick_copy_stream, _prep_data_for_correlation)


Logger = logging.getLogger(__name__)

# FDSN_CLIENT_ARGS = inspect.signature(Client).parameters.keys()
# Only allow a limited subset to be parsed through - this means
# that authenticated clients won't work.
FDSN_CLIENT_ARGS = {"base_url", "debug", "timeout"}


class _FDSN_parseable_Client:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @classmethod
    def from_client(cls, client):
        kwargs = {key: client.__dict__.get(key, None)
                  for key in FDSN_CLIENT_ARGS}
        return cls(**kwargs)

    def to_client(self):
        return Client(**self.kwargs)


###############################################################################
#                           Process handlers
###############################################################################


class Poison(Exception):
    """
    Exception passing within EQcorrscan

    :type value: Exception
    :param value: Exception to pass between processes
    """
    def __init__(self, value):
        """
        Poison Exception.
        """
        self.value = value

    def __repr__(self):
        return f"Poison({self.value.__repr__()})"

    def __str__(self):
        """
        >>> print(Poison(Exception('alf')))
        Poison(Exception('alf'))
        """
        return self.__repr__()


def _get_and_check(input_queue: Queue, poison_queue: Queue, step: float = 0.5):
    """
    Get from a queue and check for poison - returns Poisoned if poisoned.

    :param input_queue: Queue to get something from
    :param poison_queue: Queue to check for poison

    :return: Item from queue or Poison.
    """
    while True:
        poison = _check_for_poison(poison_queue)
        if poison:
            return poison
        if input_queue.empty():
            time.sleep(step)
        else:
            return input_queue.get_nowait()


def _check_for_poison(poison_queue: Queue) -> Union[Poison, None]:
    """
    Check if poison has been added to the queue.
    """
    Logger.debug("Checking for poison")
    try:
        poison = poison_queue.get_nowait()
    except Empty:
        return
    # Put the poison back in the queue for another process to check on
    Logger.error("Poisoned")
    poison_queue.put(poison)
    return Poison(poison)


def _wait_on_output_to_be_available(
    poison_queue: Queue,
    output_queue: Queue,
    raise_exception: bool = False,
    item=None,
    wait_warning: float = 60,
) -> Union[Poison, None]:
    """
    Wait until the output queue is not full to put something in it.

    :param poison_queue: Queue to put or containing poison
    :param output_queue:
        Output Queue to check whether we can put something in it
    :param item: Thing to put in the queue when we can
    :param raise_exception:
        Whether to raise an exception on poison (True), or pass back (False)

    :return: Poison if poisoned, or None if all okay

    .. rubric:: Example

    >>> from multiprocessing import Queue, Process
    >>> poison_queue, output_queue = Queue(), Queue(maxsize=1)
    >>> output_queue.put("Stopper")
    >>> process = Process(
    ...    target=_wait_on_output_to_be_available,
    ...    kwargs={"poison_queue": poison_queue,
    ...            "output_queue": output_queue,
    ...            "raise_exception": True,
    ...            "item": "Carry on",
    ...            "wait_warning": 5})
    >>> process.start()
    >>> time.sleep(7)
    >>> poison_queue.put(
    ...    Poison(Exception("cyanide")))
    >>> time.sleep(7)
    >>> process.is_alive()
    False
    >>> process.join()
    """
    killed = _check_for_poison(poison_queue)
    # Wait until output queue is empty to limit rate and memory use
    tic = default_timer()
    while output_queue.full():
        # Keep on checking while we wait
        killed = _check_for_poison(poison_queue)
        if killed:
            break
        waited = default_timer() - tic
        if waited > wait_warning:
            Logger.debug("Waiting for output_queue to not be full")
            tic = default_timer()
    if not killed and item:
        output_queue.put_nowait(item)
    elif killed and raise_exception:
        raise killed
    return killed


def _get_detection_stream(
    template_channel_ids: List[tuple],
    client,
    input_time_queue: Queue,
    retries: int,
    min_gap: float,
    buff: float,
    output_filename_queue: Queue,
    poison_queue: Queue,
    temp_stream_dir: str,
    full_stream_dir: str = None,
    pre_process: bool = False,
    parallel_process: bool = True,
    process_cores: int = None,
    daylong: bool = False,
    overlap: Union[str, float] = "calculate",
    ignore_length: bool = False,
    ignore_bad_data: bool = False,
    filt_order: int = None,
    highcut: float = None,
    lowcut: float = None,
    samp_rate: float = None,
    process_length: float = None,
):
    """
    Get a stream to be used for detection from a client for a time period.

    This function is designed to be run continuously within a Process and will
    only stop when the next item in the input_time_queue is None.

    This function uses .get_waveforms_bulk to get a Stream from a Client.
    The specific time period to get data for is read from the input_time_queue.
    Once the data have been loaded into memory from the Client, this function
    then processes that Stream according to the processing parameters passed
    as arguments to this function. Finally, this function writes the processed
    Stream to disk in a temporary stream directory and puts the filename
    for that stream in the output_filename_queue.

    Optionally, an unprocessed version of the stream can be written to the
    full_stream_dir directory to later provide an unprocessed copy of the
    raw data. This can be helpful if data downloading is slow and the stream
    is required for subsequent processing.

    :param template_channel_ids:
        Iterable of (network, station, location, channel) tuples to get data
        for. Wildcards may be used if accepted by the client.
    :param client:
        Client-like object with at least a .get_waveforms_bulk method.
    :param input_time_queue:
        Queue of (starttime, endtime) tuples of UTCDateTimes to get data
        between.
    :param retries: See core.match_filter.tribe.client_detect
    :param min_gap: See core.match_filter.tribe.client_detect
    :param buff:
        Length to pad downloaded data by - some clients do not provide all
        data requested.
    :param output_filename_queue:
        Queue to put filenames of written streams into
    :param poison_queue:
        Queue to check for poison, or put poison into if something goes awry
    :param temp_stream_dir:
        Directory to write processed streams to.
    :param full_stream_dir:
        Directory to save unprocessed streams to. If None, will not be used.
    :param pre_process: Whether to run pre-processing or not.
    :param parallel_process:
        Whether to process data in parallel (uses multi-threading)
    :param process_cores:
        Maximum number of cores to use for parallel processing
    :param daylong: See utils.pre_processing.multi_process
    :param overlap: See core.match_filter.tribe.detect
    :param ignore_length: See utils.pre_processing.multi_process
    :param ignore_bad_data: See utils.pre_processing.multi_process
    :param filt_order: See utils.pre_processing.multi_process
    :param highcut: See utils.pre_processing.multi_process
    :param lowcut: See utils.pre_processing.multi_process
    :param samp_rate: See utils.pre_processing.multi_process
    :param process_length: See utils.pre_processing.multi_process
    """
    if isinstance(client, _FDSN_parseable_Client):
        client = client.to_client()
    while True:
        killed = _wait_on_output_to_be_available(
            poison_queue=poison_queue, output_queue=output_filename_queue,
            item=False)
        if killed:
            Logger.error("Killed")
            break
        try:
            next_times = _get_and_check(input_time_queue, poison_queue)
            if next_times is None:
                break
            if isinstance(next_times, Poison):
                Logger.error("Killed")
                break
            starttime, endtime = next_times

            st = _download_st(
                starttime=starttime, endtime=endtime, buff=buff,
                min_gap=min_gap, template_channel_ids=template_channel_ids,
                client=client, retries=retries)
            if len(st) == 0:
                Logger.warning(f"No suitable data between {starttime} "
                               f"and {endtime}, skipping")
                continue
            Logger.info(f"Downloaded stream of {len(st)} traces:")
            for tr in st:
                Logger.info(tr)
            # Try to reduce memory consumption by getting rid of st if we can
            if full_stream_dir:
                for tr in st:
                    tr.split().write(os.path.join(
                        full_stream_dir,
                        f"full_trace_{tr.id}_"
                        f"{tr.stats.starttime.strftime('%y-%m-%dT%H-%M-%S')}"
                        f".ms"), format="MSEED")
            if not pre_process:
                st_chunks = [st]
            else:
                template_ids = set(['.'.join(sid)
                                    for sid in template_channel_ids])
                # Group_process copies the stream.
                st_chunks = _pre_process(
                    st=st, template_ids=template_ids, pre_processed=False,
                    filt_order=filt_order, highcut=highcut,
                    lowcut=lowcut, samp_rate=samp_rate,
                    process_length=process_length,
                    parallel=parallel_process, cores=process_cores,
                    daylong=daylong, ignore_length=ignore_length,
                    overlap=overlap, ignore_bad_data=ignore_bad_data)
                # We don't need to hold on to st!
                del st
            for chunk in st_chunks:
                Logger.info(f"After processing stream has {len(chunk)} traces:")
                for tr in chunk:
                    Logger.info(tr)
                if not os.path.isdir(temp_stream_dir):
                    os.makedirs(temp_stream_dir)
                chunk_file = os.path.join(
                    temp_stream_dir,
                    f"chunk_{len(chunk)}_"
                    f"{chunk[0].stats.starttime.strftime('%Y-%m-%dT%H-%M-%S')}"
                    f"_{os.getpid()}.pkl")
                # Add PID to cope with multiple instances operating at once
                _pickle_stream(chunk, chunk_file)
                # Wait for output queue to be ready
                _wait_on_output_to_be_available(
                    poison_queue=poison_queue,
                    output_queue=output_filename_queue,
                    item=chunk_file, raise_exception=True)
                del chunk
        except Exception as e:
            Logger.error(f"Caught exception {e} in downloader")
            poison_queue.put(Poison(e))
            traceback.print_tb(e.__traceback__)
            break
    # Wait for output queue to be ready
    killed = _wait_on_output_to_be_available(
        poison_queue=poison_queue,
        output_queue=output_filename_queue,
        raise_exception=False)
    if killed:
        poison_queue.put_nowait(killed)
    else:
        output_filename_queue.put(None)
    return


def _pre_processor(
    input_stream_queue: Queue,
    temp_stream_dir: str,
    template_ids: set,
    pre_processed: bool,
    filt_order: int,
    highcut: float,
    lowcut: float,
    samp_rate: float,
    process_length: float,
    parallel: bool,
    cores: int,
    daylong: bool,
    ignore_length: bool,
    overlap: float,
    ignore_bad_data: bool,
    output_filename_queue: Queue,
    poison_queue: Queue,
):
    """
    Consume a queue of input streams and process those streams.

    This function is designed to be run continuously within a Process and will
    only stop when the next item in the input_stream_queue is None.

    This function consumes streams from the input_stream_queue and processes
    them using utils.pre_processing functions. Processed streams are written
    out to the temp_stream_dir and the filenames are produced in the
    output_filename_queue.

    :param input_stream_queue:
        Input Queue to consume streams from.
    :param temp_stream_dir: Directory to write processed streams to.
    :param template_ids:
        Iterable of seed ids in the template set. Only channels matching these
        seed ids will be retained.
    :param pre_processed: See core.match_filter.tribe.detect
    :param filt_order: See utils.pre_processing.multi_process
    :param highcut: See utils.pre_processing.multi_process
    :param lowcut: See utils.pre_processing.multi_process
    :param samp_rate: See utils.pre_processing.multi_process
    :param process_length: See utils.pre_processing.multi_process
    :param parallel: See utils.pre_processing.multi_process
    :param cores: See utils.pre_processing.multi_process
    :param daylong: See utils.pre_processing.multi_process
    :param ignore_length: See utils.pre_processing.multi_process
    :param overlap: See core.match_filter.tribe.detect
    :param ignore_bad_data: See utils.pre_processing.multi_process
    :param output_filename_queue:
        Queue to put filenames of written streams into
    :param poison_queue:
         Queue to check for poison, or put poison into if something goes awry
    """
    while True:
        killed = _check_for_poison(poison_queue)
        if killed:
            break
        Logger.debug("Getting stream from queue")
        st = _get_and_check(input_stream_queue, poison_queue)
        if st is None:
            Logger.info("Ran out of streams, stopping processing")
            break
        elif isinstance(st, Poison):
            Logger.error("Killed")
            break
        if len(st) == 0:
            break
        Logger.info(f"Processing stream:\n{st}")

        # Process stream
        try:
            st_chunks = _pre_process(
                st, template_ids, pre_processed, filt_order, highcut, lowcut,
                samp_rate, process_length, parallel, cores, daylong,
                ignore_length, ignore_bad_data, overlap)
            for chunk in st_chunks:
                if not os.path.isdir(temp_stream_dir):
                    os.makedirs(temp_stream_dir)
                chunk_file = os.path.join(
                    temp_stream_dir,
                    f"chunk_{len(chunk)}_"
                    f"{chunk[0].stats.starttime.strftime('%Y-%m-%dT%H-%M-%S')}"
                    f"_{os.getpid()}.pkl")
                # Add PID to cope with multiple instances operating at once
                _pickle_stream(chunk, chunk_file)
                # Wait for output queue to be ready
                _wait_on_output_to_be_available(
                    poison_queue=poison_queue,
                    output_queue=output_filename_queue,
                    item=chunk_file, raise_exception=True)
                del chunk
        except Exception as e:
            Logger.error(
                f"Caught exception in processor:\n {e}")
            poison_queue.put_nowait(Poison(e))
            traceback.print_tb(e.__traceback__)
    # Wait for output queue to be ready
    killed = _wait_on_output_to_be_available(
        poison_queue=poison_queue,
        output_queue=output_filename_queue,
        raise_exception=False)
    if killed:
        poison_queue.put_nowait(killed)
    else:
        output_filename_queue.put_nowait(None)
    return


def _prepper(
    input_stream_filename_queue: Queue,
    templates: Union[List, dict],
    group_size: int,
    groups: Iterable[Iterable[str]],
    output_queue: Queue,
    poison_queue: Queue,
    xcorr_func: str = None,
):
    """
    Prepare templates and stream for correlation.

    This function is designed to be run continuously within a Process and will
    only stop when the next item in the input_stream_queue is None.

    This function prepares (reshapes into numpy arrays) templates and streams
    and ensures that the data are suitable for the cross-correlation function
    specified.

    :param input_stream_filename_queue:
        Input Queue to consume stream_filenames from.
    :param templates:
        Either (a) a list of Template objects, or (b) a dictionary of pickled
        template filenames, keyed by template name.
    :param group_size:
        See core.match_filter.tribe.detect
    :param groups:
        Iterable of groups, where each group is an iterable of the template
        names in that group.
    :param output_queue:
        Queue to produce inputs for correlation to.
    :param poison_queue:
        Queue to check for poison, or put poison into if something goes awry
    :param xcorr_func:
        Name of correlation function backend to be used.
    """
    if isinstance(templates, dict):
        # We have been passed a db of template files on disk
        Logger.info("Deserializing templates from disk")
        try:
            templates = _read_template_db(templates)
        except Exception as e:
            Logger.error(f"Could not read from db due to {e}")
            poison_queue.put_nowait(Poison(e))
            return

    while True:
        killed = _check_for_poison(poison_queue)
        if killed:
            Logger.info("Killed in prepper")
            break
        Logger.info("Getting stream from queue")
        st_file = _get_and_check(input_stream_filename_queue, poison_queue)
        if st_file is None:
            Logger.info("Got None for stream, prepper complete")
            break
        elif isinstance(st_file, Poison):
            Logger.error("Killed")
            break
        if isinstance(st_file, Stream):
            Logger.info("Stream provided")
            st = st_file
            # Write temporary cache of file
            st_file = tempfile.NamedTemporaryFile().name
            Logger.info(f"Writing temporary stream file to {st_file}")
            try:
                _pickle_stream(st, st_file)
            except Exception as e:
                Logger.error(
                    f"Could not write temporary file {st_file} due to {e}")
                poison_queue.put_nowait(Poison(e))
                break
        Logger.info(f"Reading stream from {st_file}")
        try:
            st = _unpickle_stream(st_file)
        except Exception as e:
            Logger.error(f"Error reading {st_file}: {e}")
            poison_queue.put_nowait(Poison(e))
            break
        st_sids = {tr.id for tr in st}
        if len(st_sids) < len(st):
            _sids = [tr.id for tr in st]
            _duplicate_sids = {
                sid for sid in st_sids if _sids.count(sid) > 1}
            poison_queue.put_nowait(Poison(NotImplementedError(
                f"Multiple channels in continuous data for "
                f"{', '.join(_duplicate_sids)}")))
            break
        # Do the grouping for this stream
        Logger.info(f"Grouping {len(templates)} templates into groups "
                    f"of {group_size} templates")
        try:
            template_groups = _group(sids=st_sids, templates=templates,
                                     group_size=group_size, groups=groups)
        except Exception as e:
            Logger.error(e)
            poison_queue.put_nowait(Poison(e))
            break
        Logger.info(f"Grouped into {len(template_groups)} groups")
        for i, template_group in enumerate(template_groups):
            killed = _check_for_poison(poison_queue)
            if killed:
                break
            try:
                template_streams = [
                    _quick_copy_stream(t.st) for t in template_group]
                template_names = [t.name for t in template_group]

                # template_names, templates = zip(*template_group)
                Logger.info(
                    f"Prepping {len(template_streams)} "
                    f"templates for correlation")
                # We can just load in a fresh copy of the stream!
                _st, template_streams, template_names = \
                    _prep_data_for_correlation(
                        stream=_unpickle_stream(st_file).merge(),
                        templates=template_streams,
                        template_names=template_names)
                if len(_st) == 0:
                    Logger.error(
                        f"No traces returned from correlation prep: {_st}")
                    continue
                starttime = _st[0].stats.starttime

                if xcorr_func in (None, "fmf", "fftw"):
                    array_dict_tuple = _get_array_dicts(
                        template_streams, _st, stack=True)
                    stream_dict, template_dict, pad_dict, \
                        seed_ids = array_dict_tuple
                    if xcorr_func == "fmf":
                        Logger.info("Prepping data for FMF")
                        # Work out used channels here
                        tr_chans = np.array(
                            [~np.isnan(template_dict[seed_id]).any(axis=1)
                             for seed_id in seed_ids])
                        no_chans = np.sum(np.array(tr_chans).astype(int),
                                          axis=0)
                        chans = [[] for _i in range(len(templates))]
                        for seed_id, tr_chan in zip(seed_ids, tr_chans):
                            for chan, state in zip(chans, tr_chan):
                                if state:
                                    chan.append((seed_id.split('.')[1],
                                                 seed_id.split('.')[-1].split(
                                                     '_')[0]))
                        # Reshape
                        t_arr, d_arr, weights, pads = _fmf_reshape(
                            template_dict=template_dict,
                            stream_dict=stream_dict,
                            pad_dict=pad_dict, seed_ids=seed_ids)
                        # Stabilise
                        t_arr, d_arr, multipliers = _fmf_stabilisation(
                            template_arr=t_arr, data_arr=d_arr)
                        # Wait for output queue to be ready
                        _wait_on_output_to_be_available(
                            poison_queue=poison_queue,
                            output_queue=output_queue,
                            item=(starttime, i, d_arr, template_names, t_arr,
                                  weights, pads, chans, no_chans),
                            raise_exception=True)
                    else:
                        Logger.info("Prepping data for FFTW")
                        # Wait for output queue to be ready
                        killed = _wait_on_output_to_be_available(
                            poison_queue=poison_queue,
                            output_queue=output_queue,
                            item=(starttime, i, stream_dict, template_names,
                                  template_dict, pad_dict, seed_ids),
                            raise_exception=True)
                else:
                    Logger.info("Prepping data for standard correlation")
                    # Wait for output queue to be ready
                    killed = _wait_on_output_to_be_available(
                        poison_queue=poison_queue, output_queue=output_queue,
                        item=(starttime, i, _st, template_names,
                              template_streams),
                        raise_exception=True)
            except Exception as e:
                Logger.error(f"Caught exception in Prepper: {e}")
                traceback.print_tb(e.__traceback__)
                poison_queue.put_nowait(Poison(e))
            i += 1
        Logger.info(f"Removing temporary {st_file}")
        os.remove(st_file)
    # Wait for output queue to be ready
    killed = _wait_on_output_to_be_available(
        poison_queue=poison_queue, output_queue=output_queue,
        raise_exception=False)
    if killed:
        poison_queue.put_nowait(killed)
    else:
        output_queue.put_nowait(None)
    return


def _make_detections(
    input_queue: Queue,
    delta: float,
    templates: Union[List, dict],
    threshold: float,
    threshold_type: str,
    save_progress: bool,
    output_queue: Queue,
    poison_queue: Queue,
):
    """
    Construct Detection objects from sparse detection information.

    This function is designed to be run continuously within a Process and will
    only stop when the next item in the input_queue is None.

    :param input_queue:
        Queue of (starttime, peaks, thresholds, no_channels, channels,
        template_names). Detections are made within `peaks`.
    :param delta:
        Sample rate of peaks to detect within in Hz
    :param templates:
        Template objects included in input_queue
    :param threshold:
        Overall threshold
    :param threshold_type:
        Overall threshold type
    :param save_progress:
        Whether to save progress or not: If true, individual Party files will
        be written each time this is run.
    :param output_queue:
        Queue of output Party filenames.
    :param poison_queue:
        Queue to check for poison, or put poison into if something goes awry
    """
    chunk_id = 0
    while True:
        killed = _check_for_poison(poison_queue)
        if killed:
            break
        try:
            next_item = _get_and_check(input_queue, poison_queue)
            if next_item is None:
                Logger.info("_make_detections got None, stopping")
                break
            elif isinstance(next_item, Poison):
                Logger.error("Killed")
                break
            starttime, all_peaks, thresholds, no_chans, \
                chans, template_names = next_item
            detections = _detect(
                template_names=template_names, all_peaks=all_peaks,
                starttime=starttime, delta=delta, no_chans=no_chans,
                chans=chans, thresholds=thresholds)
            Logger.info(f"Built {len(detections)}")
            chunk_file = _make_party(
                detections=detections, threshold=threshold,
                threshold_type=threshold_type, templates=templates,
                chunk_start=starttime, chunk_id=chunk_id,
                save_progress=save_progress)
            chunk_id += 1
            output_queue.put_nowait(chunk_file)
        except Exception as e:
            Logger.error(
                f"Caught exception in detector:\n {e}")
            traceback.print_tb(e.__traceback__)
            poison_queue.put_nowait(Poison(e))
    output_queue.put_nowait(None)
    return


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Utilities module whose functions are designed to do the basic processing of
the data using obspy modules (which also rely on scipy and numpy).

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import os

import numpy as np
import logging
import copy

from collections import Counter, defaultdict
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache, partial
from scipy.signal import iirfilter, sosfilt, zpk2sos

from obspy import Stream, Trace, UTCDateTime
from obspy.core.trace import Stats


Logger = logging.getLogger(__name__)


def _check_daylong(data, threshold=0.5):
    """
    Check data continuity.

    Check to see that the day is more than threshold of zeros, if it is
    then the resampling will hate it.

    :type data: np.array
    :param data: Data from Trace to check if the data are okay.
    :type threshold: float
    :param threshold: Fraction of data to accept as zeros.

    :return quality (simply good or bad)
    :rtype: bool

    .. rubric:: Example

    >>> from obspy import read
    >>> from eqcorrscan.utils.pre_processing import _check_daylong
    >>> # Get the path to the test data
    >>> import eqcorrscan
    >>> import os
    >>> TEST_PATH = os.path.dirname(eqcorrscan.__file__) + '/tests/test_data'
    >>> st = read(TEST_PATH + '/WAV/TEST_/' +
    ...           '2013-09-01-0410-35.DFDPC_024_00')
    >>> _check_daylong(st[0].data)
    True
    >>> zeroed_data = st[0].copy().data
    >>> zeroed_data[0:9100] = np.zeros(9100)
    >>> _check_daylong(zeroed_data)
    False
    """
    return np.nonzero(data)[0].shape[0] >= threshold * data.shape[0]


def _simple_qc(st, max_workers=None, chunksize=1):
    """
    Multithreaded simple QC of data.

    :param st: Stream of data to check
    :type st: obspy.core.Stream
    :param max_workers: Maximum number of threads to use
    :type max_workers: int
    :param chunksize: Number of traces to process per thread
    :type chunksize: int

    :return: dict of {tr.id: quality} where quality is bool
    """
    qual = dict()
    with ThreadPoolExecutor(max_workers) as executor:
        for tr, _qual in zip(st, executor.map(
                _check_daylong, (tr.data for tr in st), chunksize=chunksize)):
            qual[tr.id] = _qual
    return qual


def _sanitize_length(st, starttime=None, endtime=None, daylong=False):
    """
    Check length and work out start, end, length and trimming criteria

    :param st: Stream to check
    :type st: obspy.core.Stream
    :param starttime: Desired starttime - if None, will be evaluated from data
    :type starttime: obspy.core.UTCDateTime
    :param endtime: DEsired endtime - can be None
    :type endtime: obspy.core.UTCDateTime
    :param daylong: Whether data should be one-day long.
    :type daylong: bool

    :return: obspy.core.Stream, length[float], clip[bool],
        starttime[obspy.core.UTCDateTime]
    """
    length, clip = None, False

    if daylong:
        length, clip = 86400, True
        # Set the start-time to a day start - cope with
        if starttime is None:
            startdates = []
            for tr in st:
                if abs(tr.stats.starttime - (UTCDateTime(
                        tr.stats.starttime.date) + 86400)) < tr.stats.delta:
                    # If the trace starts within 1 sample of the next day,
                    # use the next day as the startdate
                    startdates.append((tr.stats.starttime + 86400).date)
                    Logger.warning(
                        f'{tr.id} starts within 1 sample of the next day, '
                        f'using this time {(tr.stats.starttime + 86400).date}')
                else:
                    startdates.append(tr.stats.starttime.date)
            # Check that all traces start on the same date...
            if not len(set(startdates)) == 1:
                raise NotImplementedError('Traces start on different days')
            starttime = UTCDateTime(startdates[0])
    else:
        if starttime is not None and endtime is not None:
            for tr in st:
                Logger.info(
                    f"Trimming {tr.id} between {starttime} and {endtime}")
                tr.trim(starttime, endtime)
                if len(tr.data) == ((endtime - starttime) *
                                    tr.stats.sampling_rate) + 1:
                    Logger.info(f"{tr.id} is overlength dropping first sample")
                    tr.data = tr.data[1:len(tr.data)]
                    # TODO: this should adjust the start-time
                    # tr.stats.starttime += tr.stats.delta
            length = endtime - starttime
            clip = True
        elif starttime:
            for tr in st:
                tr.trim(starttime=starttime)
        elif endtime:
            for tr in st:
                tr.trim(endtime=endtime)
    return st, length, clip, starttime


@lru_cache(maxsize=5)
def _get_window(window, npts):
    """ Get window for resampling stabilisation. """
    from scipy.signal import get_window
    return np.fft.ifftshift(get_window(window, npts))


def multi_process(st, lowcut, highcut, filt_order, samp_rate, parallel=False,
                  num_cores=False, starttime=None, endtime=None,
                  daylong=False, seisan_chan_names=False, fill_gaps=True,
                  ignore_length=False, ignore_bad_data=False):
    """
    Apply standardised processing workflow to data for matched-filtering

    Steps:

        #. Check length and continuity of data meets user-defined criteria
        #. Fill remaining gaps in data with zeros and record gap positions
        #. Detrend data (using a simple linear detrend to set start and
           end to 0)
        #. Pad data to length
        #. Resample in the frequency domain
        #. Detrend data (using a simple linear detrend to set start and
           end to 0)
        #. Zerophase Butterworth filter
        #. Re-check length
        #. Re-apply zero-padding to gap locations recording in step 2 to remove
           filtering and resampling artefacts

    :param st: Stream to process
    :type st: obspy.core.Stream
    :param lowcut:
        Lowcut of butterworth filter in Hz. If set to None and highcut is
        given a highpass filter will be applied. If both lowcut and highcut
        are given, a bandpass filter will be applied. If lowcut and highcut
        are both None, no filtering will be applied.
    :type lowcut: float
    :param highcut:
        Highcut of butterworth filter in Hz. If set to None and lowcut is
        given a lowpass filter will be applied. If both lowcut and highcut
        are given, a bandpass filter will be applied. If lowcut and highcut
        are both None, no filtering will be applied.
    :type highcut: float
    :param filt_order: Filter order
    :type filt_order: int
    :param samp_rate: Desired sample rate of output data in Hz
    :type samp_rate: float
    :param parallel: Whether to process data in parallel (uses multi-threading)
    :type parallel: bool
    :param num_cores: Maximum number of cores to use for parallel processing
    :type num_cores: int
    :param starttime: Desired starttime of data
    :type starttime: obspy.core.UTCDateTime
    :param endtime: Desired endtime of data
    :type endtime: obspy.core.UTCDateTime
    :param daylong:
        Whether data should be considered to be one-day long. Setting this will
        assume that your data should start as close to the start of a day
        as possible given the sampling.
    :type daylong: bool
    :param seisan_chan_names:
        Whether to convert channel names to two-char seisan channel names
    :type seisan_chan_names: bool
    :param fill_gaps: Whether to fill-gaps in the data
    :type fill_gaps: bool
    :param ignore_length:
        Whether to ignore data that are not long enough.
    :type ignore_length: bool
    :param ignore_bad_data: Whether to ignore data that are excessively gappy
    :type ignore_bad_data: bool

    :return: Processed stream as obspy.core.Stream

    :Note: Works in place on your data, copy before giving to this function if
           you want to reuse your input data.
    """
    if isinstance(st, Trace):
        tracein = True
        st = Stream(st)
    else:
        tracein = False
    # Add sanity check for filter
    if highcut and highcut >= 0.5 * samp_rate:
        raise IOError('Highcut must be lower than the Nyquist')
    if highcut and lowcut:
        assert lowcut < highcut, f"Lowcut: {lowcut} above highcut: {highcut}"

    # Allow datetimes for starttime and endtime
    if starttime and not isinstance(starttime, UTCDateTime):
        starttime = UTCDateTime(starttime)
    if starttime is False:
        starttime = None
    if endtime and not isinstance(endtime, UTCDateTime):
        endtime = UTCDateTime(endtime)
    if endtime is False:
        endtime = None

    # Make sensible choices about workers and chunk sizes
    if parallel:
        if not num_cores:
            # We don't want to over-specify threads, we don't have IO
            # bound tasks
            max_workers = min(len(st), os.cpu_count())
        else:
            max_workers = min(len(st), num_cores)
    else:
        max_workers = 1
    chunksize = len(st) // max_workers

    st, length, clip, starttime = _sanitize_length(
        st=st, starttime=starttime, endtime=endtime, daylong=daylong)

    for tr in st:
        if len(tr.data) == 0:
            st.remove(tr)
            Logger.warning('No data for {0} after trim'.format(tr.id))

    # Do work
    # 0. Enforce double-preccision floats for this work
    for tr in st:
        if not tr.data.dtype == np.float64:
            Logger.debug(f"Converting {tr.id} to double precision")
            tr.data = tr.data.astype(np.float64)
    # 1. Fill gaps and keep track of them
    gappy = {tr.id: False for tr in st}
    gaps = dict()
    for i, tr in enumerate(st):
        if isinstance(tr.data, np.ma.MaskedArray):
            gappy[tr.id] = True
            gaps[tr.id], tr = _fill_gaps(tr)
            st[i] = tr

    # 2. Check for zeros and cope with bad data
    # ~ 4x speedup for 50 100 Hz daylong traces on 12 threads
    qual = _simple_qc(st, max_workers=max_workers, chunksize=chunksize)
    for trace_id, _qual in qual.items():
        if not _qual:
            msg = ("Data have more zeros than actual data, please check the "
                   f"raw data set-up and manually sort it: {tr.id}")
            if not ignore_bad_data:
                raise ValueError(msg)
            else:
                # Remove bad traces from the stream
                try:
                    st.remove(st.select(id=trace_id))
                except ValueError:
                    Logger.info(
                        f"{trace_id} not found in {set(tr.id for tr in st)},"
                        f" ignoring")

    # 3. Detrend
    # ~ 2x speedup for 50 100 Hz daylong traces on 12 threads
    st = _multi_detrend(st, max_workers=max_workers, chunksize=chunksize)

    # 4. Check length and pad to length
    padded = {tr.id: (0., 0.) for tr in st}
    if clip:
        st.trim(starttime, starttime + length, nearest_sample=True)
        # Indexing because we are going to overwrite traces
        for i, _ in enumerate(st):
            if float(st[i].stats.npts / st[i].stats.sampling_rate) != length:
                Logger.info(
                    'Data for {0} are not long-enough, will zero pad'.format(
                        st[i].id))
                st[i], padded[st[i].id] = _length_check(
                    st[i], starttime=starttime, length=length,
                    ignore_length=ignore_length,
                    ignore_bad_data=ignore_bad_data)
        # Remove None traces that might be returned from length checking
        st.traces = [tr for tr in st if tr is not None]

    # Check that we actually still have some data
    if not _stream_has_data(st):
        if tracein:
            return st[0]
        return st

    # 5. Resample
    # ~ 3.25x speedup for 50 100 Hz daylong traces on 12 threads
    st = _multi_resample(
        st, sampling_rate=samp_rate, max_workers=max_workers,
        chunksize=chunksize)

    # Detrend again before filtering
    st = _multi_detrend(st, max_workers=max_workers, chunksize=chunksize)

    # 6. Filter
    # ~3.25x speedup for 50 100 Hz daylong traces on 12 threads
    st = _multi_filter(
        st, highcut=highcut, lowcut=lowcut, filt_order=filt_order,
        max_workers=max_workers, chunksize=chunksize)

    # 7. Reapply zeros after processing from 4
    for tr in st:
        # Pads default to (0., 0.), pads should only ever be positive.
        if sum(padded[tr.id]) == 0:
            continue
        Logger.debug("Reapplying zero pads post processing")
        Logger.debug(str(tr))
        pre_pad = np.zeros(int(padded[tr.id][0] * tr.stats.sampling_rate))
        post_pad = np.zeros(int(padded[tr.id][1] * tr.stats.sampling_rate))
        pre_pad_len = len(pre_pad)
        post_pad_len = len(post_pad)
        Logger.debug(
            f"Taking only valid data between {pre_pad_len} and "
            f"{tr.stats.npts - post_pad_len} samples")
        # Re-apply the pads, taking only the data section that was valid
        tr.data = np.concatenate(
            [pre_pad, tr.data[pre_pad_len: len(tr.data) - post_pad_len],
             post_pad])
        Logger.debug(str(tr))

    # 8. Recheck length
    for tr in st:
        if float(tr.stats.npts * tr.stats.delta) != length and clip:
            Logger.info(f'Data for {tr.id} are not of required length, will '
                        f'zero pad')
            # Use obspy's trim function with zero padding
            tr = tr.trim(starttime, starttime + length, pad=True, fill_value=0,
                         nearest_sample=True)
            # If there is one sample too many after this remove the last one
            # by convention
            if len(tr.data) == (length * tr.stats.sampling_rate) + 1:
                tr.data = tr.data[1:len(tr.data)]
            if abs((tr.stats.sampling_rate * length) -
                   tr.stats.npts) > tr.stats.delta:
                raise ValueError('Data are not required length for ' +
                                 tr.stats.station + '.' + tr.stats.channel)

    # 9. Re-insert gaps from 1
    for i, tr in enumerate(st):
        if gappy[tr.id]:
            st[i] = _zero_pad_gaps(tr, gaps[tr.id], fill_gaps=fill_gaps)

    # 10. Clean up
    for tr in st:
        if len(tr.data) == 0:
            st.remove(tr)

    # 11. Account for seisan channel naming
    if seisan_chan_names:
        for tr in st:
            tr.stats.channel = tr.stats.channel[0] + tr.stats.channel[-1]

    if tracein:
        st.merge()
        return st[0]

    return st


@lru_cache(maxsize=50)
def _empty_trace(
        network,
        station,
        location,
        channel,
        starttime,
        sampling_rate
):
    """
    Generate an empty trace with a basic header matching the input trace

    :param network: Network code
    :type network: str
    :param station: Station code
    :type station: str
    :param location: Location code
    :type location: str
    :param channel: Channel code
    :type channel: str
    :param starttime: Start time of trace as datetime (NOT UTCDateTime)
    :type starttime: datetime.datetime
    :param sampling_rate: Sampling rate of data
    :type sampling_rate: float

    :returns: trace
    """
    bad_trace = Trace(
        data=np.array([]), header={
            "station": station, "channel": channel,
            "network": network, "location": location,
            "starttime": starttime,
            "sampling_rate": sampling_rate})
    return bad_trace


def _stream_has_data(st):
    return sum(tr.stats.npts for tr in st) > 0


def _length_check(tr, starttime, length, ignore_length, ignore_bad_data):
    """
    Check that a trace meets the length requirements specified.

    Data are padded if needed to meet the length requirement.

    :param tr: Trace to check
    :type tr: obspy.core.Trace
    :param starttime: Desired starttime of data
    :type starttime: obspy.core.UTCDateTime
    :param length: Length in seconds required for data
    :type length: float
    :param ignore_length:
        Whether to ignore data that do not meet length criteria
    :type ignore_length: bool
    :param ignore_bad_data:
        Whether to ignore data that do not meet gappiness criteria
    :type ignore_bad_data: bool

    :return: obspy.core.Trace that meets criteria
    """
    trace_length = tr.stats.endtime - tr.stats.starttime
    if trace_length < 0.8 * length and not ignore_length:
        msg = f"Data for {tr.id} is {trace_length:.2f} seconds "\
              f"long, which is less than 80 percent of the desired "\
              f"length ({length} seconds), will not pad"
        if not ignore_bad_data:
            raise NotImplementedError(msg)
        else:
            Logger.warning(msg)
            return _empty_trace(tr.stats.network, tr.stats.station,
                                tr.stats.location, tr.stats.channel,
                                tr.stats.starttime.datetime,
                                tr.stats.sampling_rate), (0., 0.)
    # trim, then calculate length of any pads required
    pre_pad_secs = tr.stats.starttime - starttime
    post_pad_secs = (starttime + length) - tr.stats.endtime
    if pre_pad_secs > 0 or post_pad_secs > 0:
        pre_pad = np.zeros(int(pre_pad_secs * tr.stats.sampling_rate))
        post_pad = np.zeros(int(post_pad_secs * tr.stats.sampling_rate))
        Logger.debug(str(tr))
        Logger.info(
            f"Padding to length with {pre_pad_secs} s before "
            f"and {post_pad_secs} s at end")
        tr.data = np.concatenate([pre_pad, tr.data, post_pad])
        # Use this rather than the expected pad because of rounding samples
        tr.stats.starttime -= len(pre_pad) * tr.stats.delta
        Logger.debug(str(tr))
    # If there is one sample too many after this remove the first one
    # by convention
    if tr.stats.npts == (length * tr.stats.sampling_rate) + 1:
        tr.data = tr.data[1:len(tr.data)]
    # Cope with time precision.
    if abs((tr.stats.sampling_rate * length) -
           tr.stats.npts) > tr.stats.delta:
        msg = (f"Data sampling-rate * length ({tr.stats.sampling_rate} *"
               f" {length} = {tr.stats.sampling_rate * length}) does not "
               f"match number of samples ({tr.stats.npts}) for {tr.id}")
        if not ignore_bad_data:
            raise ValueError(msg)
        else:
            Logger.warning(msg)
            return _empty_trace(tr.stats.network, tr.stats.station,
                                tr.stats.location, tr.stats.channel,
                                tr.stats.starttime.datetime,
                                tr.stats.sampling_rate), (0., 0.)
    Logger.debug(
        f'I now have {tr.stats.npts} data points after enforcing length')
    return tr, (pre_pad_secs, post_pad_secs)


def _multi_filter(st, highcut, lowcut, filt_order, max_workers=None,
                  chunksize=1):
    """
    Multithreaded zero-phase butterworth filtering of multi-channel data.

    :param st: Stream to filter
    :type st: obspy.core.Stream
    :param highcut: Highcut for butterworth filter in Hz
    :type highcut: float
    :param lowcut: Lowcut for butterworth filter in Hz
    :type lowcut: float
    :param filt_order: Filter order
    :type filt_order: int
    :param max_workers: Maximum number of threads to use
    :type max_workers: int
    :param chunksize: Number of traces to process per thread
    :type chunksize: int

    :return: obspy.core.Stream of filtered data
    """
    if not highcut and not lowcut:
        Logger.warning("No filters applied")
        return st
    # Require that all channels are the same sampling frequency
    samp_rate = set(tr.stats.sampling_rate for tr in st)
    assert len(samp_rate) == 1, "Different sampling rates found"
    samp_rate = samp_rate.pop()
    # Sanity check filter bounds
    if highcut:
        assert highcut * 2 < samp_rate, "Highcut must be below Nyquist"
    if highcut and lowcut:
        assert lowcut < highcut, "Lowcut must be below highcut"

    fe = 0.5 * samp_rate
    if lowcut:
        low = lowcut / fe
    if highcut:
        high = highcut / fe

    # Design filter
    if highcut and lowcut:
        z, p, k = iirfilter(
            filt_order, [low, high], btype='band',
            ftype='butter', output='zpk')
    elif highcut:
        z, p, k = iirfilter(
            filt_order, high, btype='lowpass', ftype='butter',
            output='zpk')
    elif lowcut:
        z, p, k = iirfilter(
            filt_order, low, btype='highpass', ftype='butter',
            output='zpk')

    sos = zpk2sos(z, p, k)

    _filter = partial(_zerophase_filter, sos)

    with ThreadPoolExecutor(max_workers) as executor:
        results = executor.map(
            _filter, (tr.data for tr in st), chunksize=chunksize)

    for r, tr in zip(results, st):
        tr.data = r

    return st


def _zerophase_filter(sos, data):
    """
    Simple zerophase implementation of sosfilt.

    :param sos: Second-order-series of filters
    :param data: Data to filter
    :return: filtered data
    """
    if len(data) == 0:
        Logger.debug("No data, no filtering")
        return data
    firstpass = sosfilt(sos, data)
    return sosfilt(sos, firstpass[::-1])[::-1]


def _multi_detrend(st, max_workers=None, chunksize=1):
    """
    Multithreaded detrending using simple linear detrend between
    first and last values. Follows obspy "simple" detrend.

    :param st: Stream to detrend
    :type st: obspy.core.Stream
    :param max_workers: Maximum number of threads to use
    :type max_workers: int
    :param chunksize: Number of traces to process per thread
    :type chunksize: int

    :return: obspy.core.Stream of detrended data
    """
    for tr in st:
        tr.data = np.require(tr.data, np.float64)
    with ThreadPoolExecutor(max_workers) as executor:
        results = executor.map(_detrend, (tr.data for tr in st),
                               chunksize=chunksize)
    # Ensure tasks complete
    _ = (r for r in results)
    return st


def _detrend(data):
    """
    Detrend signal simply by subtracting a line through the first and last
    point of the trace

    :param data: Data to detrend
    :type data: np.ndarray.
    :return: Nothing - works in place
    """
    # Work in double-precision.
    data = np.require(data, dtype=np.float64)
    ndat = data.shape[0]
    x1, x2 = data[0], data[-1]
    data -= x1 + np.arange(ndat, dtype=np.float64) * (
        np.float64(x2 - x1) / np.float64(ndat - 1))
    return


def _multi_resample(st, sampling_rate, max_workers=None, chunksize=1):
    """
    Threaded resampling of a stream of data to a consistent sampling-rate

    :param st: Stream to resample
    :type st: obspy.core.Stream
    :param sampling_rate: Sampling rate to resample to
    :type sampling_rate: float
    :param max_workers: Maximum number of threads to use
    :type max_workers: int
    :param chunksize: Number of traces to process per thread
    :type chunksize: int

    :return: obspy.core.Stream of resampled data
    """
    # Get the windows, and downsampling factors ahead of time
    to_resample = (
        (tr.data, tr.stats.delta,
         tr.stats.sampling_rate / float(sampling_rate),
         sampling_rate, _get_window("hann", tr.stats.npts), tr.id)
        for tr in st)
    with ThreadPoolExecutor(max_workers) as executor:
        # Unpack tuple using lambda
        results = executor.map(lambda args: _resample(*args), to_resample,
                               chunksize=chunksize)
    for r, tr in zip(results, st):
        tr.data = r
        tr.stats.sampling_rate = sampling_rate
    return st


def _resample(data, delta, factor, sampling_rate, large_w, _id):
    """
    Resample data in the frequency domain - adapted from obspy resample method

    :param data: Data to resample
    :type data: np.ndarray
    :param delta: Sample interval in seconds
    :type delta: float
    :param factor: Factor to resample by
    :type factor: float
    :param sampling_rate: Desired sampling-rate
    :type sampling_rate: float
    :param large_w: Window to apply to spectra to stabilise resampling
    :type large_w: np.ndarray

    :return: np.ndarray of resampled data.
    """
    if factor == 1:
        # No resampling needed, don't waste time.
        return data
    # Need to work with numpy objects to release the GIL
    npts = data.shape[0]
    Logger.debug(f"Running resample for {_id} with {npts} data points")
    Logger.debug(f"{_id}: delta={delta}, factor={factor}, "
                 f"sampling_rate out={sampling_rate}")
    Logger.debug(f"Sanity check data for {_id}, start and "
                 f"end: {data[0]} -- {data[-1]}")
    Logger.debug(f"dtype for {_id}: {data.dtype}")
    if data.dtype == np.dtype('float64'):
        _floater = np.float64  # Retain double-precision
    else:
        _floater = np.float32
        # Use single-precision where possible to reduce memory
    data = data.astype(_floater)
    df = _floater(1.0) / (npts * delta)
    num = np.int32(npts / factor)
    d_large_f = _floater(1.0) / num * sampling_rate

    # Forward fft
    # x = np.fft.rfft(data)
    x = np.fft.rfft(data.view(data.dtype.newbyteorder("=")))
    # obspy for numpy 2.x
    # x = rfft(self.data.view(self.data.dtype.newbyteorder("=")))
    # Window
    x *= large_w[:npts // 2 + 1]

    # interpolate
    f = df * np.arange(0, npts // 2 + 1, dtype=np.int32)
    n_large_f = num // 2 + 1
    large_f = d_large_f * np.arange(0, n_large_f, dtype=np.int32)

    # Have to split into real and imaginary parts for interpolation.
    y = np.interp(large_f, f, np.real(x)) + (1j * np.interp(
        large_f, f, np.imag(x)))
    # Try to reduce memory before doing the ifft
    del large_f, f, x

    return np.fft.irfft(y, n=num)[0:num] * (_floater(num) / _floater(npts))


def _zero_pad_gaps(tr, gaps, fill_gaps=True):
    """
    Replace padded parts of trace with zeros.

    Will cut around gaps, detrend, then pad the gaps with zeros.

    :type tr: :class:`osbpy.core.stream.Trace`
    :param tr: A trace that has had the gaps padded
    :param gaps: List of dict of start-time and end-time as UTCDateTime objects
    :type gaps: list
    :param fill_gaps: Whether to fill gaps with zeros, or leave them as gaps
    :type fill_gaps: bool

    :return: :class:`obspy.core.stream.Trace`
    """
    start_in, end_in = (tr.stats.starttime, tr.stats.endtime)
    tr = Stream([tr])  # convert to stream to use cutout method
    for gap in gaps:
        Logger.debug(
            f"Filling gap between {gap['starttime']} and {gap['endtime']}")
        tr.cutout(gap['starttime'], gap['endtime']).merge()
    tr = tr.merge()[0]
    if fill_gaps:
        tr = tr.split()
        tr = tr.detrend()
        tr = tr.merge(fill_value=0)[0]
        # Need to check length - if a gap happened overlapping the end or start
        #  of the trace this will be lost.
        if tr.stats.starttime != start_in:
            # pad with zeros
            tr.data = np.concatenate(
                [np.zeros(int(tr.stats.starttime - start_in)), tr.data])
            tr.stats.starttime = start_in
        if tr.stats.endtime != end_in:
            tr.data = np.concatenate(
                [tr.data, np.zeros(int(end_in - tr.stats.endtime))])
    return tr


def _fill_gaps(tr):
    """
    Interpolate through gaps and work-out where gaps are.

    :param tr: Gappy trace (e.g. tr.data is np.ma.MaskedArray)
    :type tr: `obspy.core.stream.Trace`

    :return: gaps, trace, where gaps is a list of dict
    """
    tr = tr.split()
    gaps = tr.get_gaps()
    tr = tr.detrend().merge(fill_value=0)[0]
    gaps = [{'starttime': gap[4], 'endtime': gap[5]} for gap in gaps]
    if len(gaps):
        Logger.debug(f"Gaps in {tr.id}: \n\t{gaps}")
    return gaps, tr


def _group_process(filt_order, highcut, lowcut, samp_rate, process_length,
                   parallel, cores, stream, daylong,
                   ignore_length, ignore_bad_data, overlap):
    """
    Process and chunk data.

    :type parallel: bool
    :param parallel: Whether to use parallel processing or not
    :type cores: int
    :param cores: Number of cores to use, can be False to use all available.
    :type stream: :class:`obspy.core.stream.Stream`
    :param stream: Stream to process, will be left intact.
    :type daylong: bool
    :param daylong: Whether to enforce day-length files or not.
    :type ignore_length: bool
    :param ignore_length:
        If using daylong=True, then processing will try check that the data
        are there for at least 80% of the day, if you don't want this check
        (which will raise an error if too much data are missing) then set
        ignore_length=True.  This is not recommended!
    :type ignore_bad_data: bool
    :param ignore_bad_data:
        If False (default), errors will be raised if data are excessively
        gappy or are mostly zeros. If True then no error will be raised, but
        an empty trace will be returned.
    :type overlap: float
    :param overlap: Number of seconds to overlap chunks by.

    :return: list of processed streams.
    """
    processed_streams = []
    kwargs = {
        'filt_order': filt_order,
        'highcut': highcut, 'lowcut': lowcut,
        'samp_rate': samp_rate, 'parallel': parallel,
        'num_cores': cores, 'ignore_length': ignore_length,
        'ignore_bad_data': ignore_bad_data}
    # Processing always needs to be run to account for gaps - pre-process will
    # check whether filtering and resampling needs to be done.

    starttimes = sorted([tr.stats.starttime for tr in stream])
    endtimes = sorted([tr.stats.endtime for tr in stream])

    if daylong:
        if process_length != 86400:
            Logger.warning(
                f'Processing day-long data, but template was cut from '
                f'{process_length} s long data, will reduce correlations')
        process_length = 86400
        # Check that data all start on the same day, otherwise strange
        # things will happen...
        startdates = [starttime.date for starttime in starttimes]
        if not len(set(startdates)) == 1:
            Logger.warning('Data start on different days, setting to last day')
            starttime = UTCDateTime(startdates[-1])
        else:
            starttime = UTCDateTime(startdates[0])  # Can take any
    else:
        # We want to use shortproc to allow overlaps
        starttime = starttimes[0]
    endtime = endtimes[-1]
    data_len_samps = round((endtime - starttime) * samp_rate) + 1
    assert overlap < process_length, "Overlap must be less than process length"
    chunk_len_samps = (process_length - overlap) * samp_rate
    n_chunks = int(data_len_samps // chunk_len_samps)
    Logger.info(f"Splitting these data in {n_chunks} chunks")
    if n_chunks == 0:
        Logger.error('Data must be process_length or longer, not computing')
        return []

    for i in range(n_chunks):
        kwargs.update(
            {'starttime': starttime + (i * (process_length - overlap))})
        if not daylong:
            _endtime = kwargs['starttime'] + process_length
            kwargs.update({'endtime': _endtime})
        else:
            _endtime = kwargs['starttime'] + 86400

        # This is where data should be copied and only here!
        if n_chunks > 1:
            chunk_stream = _quick_copy_stream(
                stream.slice(starttime=kwargs['starttime'], endtime=_endtime))
            # Reduce memory by removing data that we don't need anymore
            stream.trim(starttime=_endtime - overlap)
        else:
            # If we only have one chunk, lets just use those data!
            chunk_stream = stream.trim(
                starttime=kwargs['starttime'], endtime=_endtime)
        Logger.info(f"Processing chunk {i} between {kwargs['starttime']} "
                    f"and {_endtime}")
        if len(chunk_stream) == 0:
            Logger.warning(
                f"No data between {kwargs['starttime']} and {_endtime}")
            continue
        # Enforce chunk npts
        for tr in chunk_stream:
            Logger.info(
                f"Enforcing {int(process_length * tr.stats.sampling_rate)} "
                f"samples for {tr.id} (had {tr.stats.npts} points)")
            tr.data = tr.data[0:int(
                process_length * tr.stats.sampling_rate)]
        _chunk_stream_lengths = {
            tr.id: tr.stats.endtime - tr.stats.starttime
            for tr in chunk_stream}
        for tr_id, chunk_length in _chunk_stream_lengths.items():
            # Remove traces that are too short.
            if not ignore_length and chunk_length <= .8 * process_length:
                tr = chunk_stream.select(id=tr_id)[0]
                chunk_stream.remove(tr)
                Logger.warning(
                    "Data chunk on {0} starting {1} and ending {2} is "
                    "below 80% of the requested length, will not use"
                    " this.".format(
                        tr.id, tr.stats.starttime, tr.stats.endtime))
        if len(chunk_stream) == 0:
            continue
        Logger.debug(
            f"Processing chunk:\n{chunk_stream.__str__(extended=True)}")
        Logger.info(f"Processing using {kwargs}")
        _processed_stream = multi_process(st=chunk_stream, **kwargs)
        # If data have more zeros then pre-processing will return a
        # trace of 0 length
        _processed_stream.traces = [
            tr for tr in _processed_stream if tr.stats.npts != 0]
        if len(_processed_stream) == 0:
            Logger.warning(
                f"Data quality insufficient between {kwargs['starttime']}"
                f" and {_endtime}")
            continue
        # Pre-processing does additional checks for zeros - we need to check
        # again whether we actually have something useful from this.
        processed_chunk_stream_lengths = [
            tr.stats.endtime - tr.stats.starttime
            for tr in _processed_stream]
        if min(processed_chunk_stream_lengths) >= .8 * process_length:
            processed_streams.append(_processed_stream)
        else:
            Logger.warning(
                f"Data quality insufficient between {kwargs['starttime']}"
                f" and {_endtime}")
            continue

    if _endtime < stream[0].stats.endtime:
        Logger.warning(
            "Last bit of data between {0} and {1} will go unused "
            "because it is shorter than a chunk of {2} s".format(
                _endtime, stream[0].stats.endtime, process_length))
    return processed_streams


def _quick_copy_trace(trace, deepcopy_data=True):
    """
    Function to quickly copy a trace. Sets values in the traces' and trace
    header's dict directly, circumventing obspy's init functions.
    Speedup: from 37 us to 12 us per trace - 3x faster

    :type trace: :class:`obspy.core.trace.Trace`
    :param trace: Stream to quickly copy
    :type deepcopy_data: bool
    :param deepcopy_data:
        Whether to deepcopy trace data (with `deepcopy_data=False` expect up to
        20 % speedup, but use only when you know that data trace contents will
        not change or affect results). Warning: do not use this option to copy
        traces with processing history or response information.
    :rtype: :class:`obspy.core.trace.Trace`
    return: trace
    """
    new_trace = Trace()
    for key, value in trace.__dict__.items():
        if key == 'stats':
            new_stats = new_trace.stats
            for key_2, value_2 in value.__dict__.items():
                if isinstance(value_2, UTCDateTime):
                    new_stats.__dict__[key_2] = UTCDateTime(
                        ns=value_2.__dict__['_UTCDateTime__ns'])
                else:
                    new_stats.__dict__[key_2] = value_2
        elif deepcopy_data:
            # data needs to be deepcopied (and anything else, to be safe)
            new_trace.__dict__[key] = copy.deepcopy(value)
        else:  # No deepcopy, e.g. for NaN-traces with no effect on results
            new_trace.__dict__[key] = value
    return new_trace


def _quick_copy_stream(stream, deepcopy_data=True):
    """
    Function to quickly copy a stream.
    Speedup for simple trace:
        from 112 us to 44 (35) us per 3-trace stream - 2.8x (3.2x) faster

    Warning: use `deepcopy_data=False` (saves extra ~20 % time) only when the
             changing the data in the stream later does not change results
             (e.g., for NaN-trace or when data array will not be changed).

    This is what takes longest (1 empty trace, total time to copy 27 us):
    copy header: 18 us (vs create new empty header: 683 ns)
    Two points that can speed up copying / creation:
        1. circumvent trace.__init__ and trace.__set_attr__ by setting value
           directly in trace's __dict__
        2. when setting trace header, circumvent that Stats(header) is called
           when header is already a Stats instance

    :type stream: :class:`obspy.core.stream.Stream`
    :param stream: Stream to quickly copy
    :type deepcopy_data: bool
    :param deepcopy_data:
        Whether to deepcopy data (with `deepcopy_data=False` expect up to 20 %
        speedup, but use only when you know that data trace contents will not
        change or affect results).

    :rtype: :class:`obspy.core.stream.Stream`
    return: stream
    """
    new_traces = list()
    for trace in stream:
        new_traces.append(
            _quick_copy_trace(trace, deepcopy_data=deepcopy_data))
    return Stream(new_traces)


def _stream_quick_select(stream, seed_id):
    """
    4x quicker selection of traces in stream by full Seed-ID. Does not support
    wildcards or selection by network/station/location/channel alone.
    """
    net, sta, loc, chan = seed_id.split('.')
    stream = Stream(
        [tr for tr in stream
         if (tr.stats.network == net and
             tr.stats.station == sta and
             tr.stats.location == loc and
             tr.stats.channel == chan)])
    return stream


def _prep_data_for_correlation(stream, templates, template_names=None,
                               force_stream_epoch=True):
    """
    Check that all channels are the same length and that all channels have data
    for both template and stream.

    Works in place on data - will cut to shortest length

    :param stream: Stream to compare data to
    :param templates:
        List of streams that will be forced to have the same channels as stream
    :param template_names:
        List of strings same length as templates
    :type force_stream_epoch: bool
    :param force_stream_epoch:
        Whether to force all channels in stream to cover the same time period

    :return: stream, templates, template_names (if template_names given)
    """
    n_templates = len(templates)
    template_samp_rates = {
        tr.stats.sampling_rate for template in templates for tr in template}
    stream_samp_rates = {tr.stats.sampling_rate for tr in stream}
    samp_rates = template_samp_rates.union(stream_samp_rates)
    assert len(samp_rates) == 1, "Sampling rates differ"
    samp_rate = samp_rates.pop()

    out_stream = Stream()

    named = True
    if template_names is None:
        named = False
        template_names = range(n_templates)

    # Work out shapes.
    stream_start = min([tr.stats.starttime for tr in stream])
    stream_end = max([tr.stats.endtime for tr in stream])
    if force_stream_epoch:
        stream_length = int(samp_rate * (stream_end - stream_start)) + 1
    else:
        stream_length = max([tr.stats.npts for tr in stream])

    template_length = {
        tr.stats.npts for template in templates for tr in template}
    assert len(template_length) == 1, "Template traces not all the same length"
    template_length = template_length.pop()

    stream_ids = {tr.id for tr in stream}

    # Need to ensure that a channel can be in the template multiple times.
    all_template_ids = [
        Counter([tr.id for tr in template]) for template in templates]
    template_ids = {
        stream_id: max(tid.get(stream_id, 0) for tid in all_template_ids)
        for stream_id in stream_ids}
    template_ids = {_id: value for _id, value in template_ids.items() if value}

    seed_ids = sorted(
        [key.split('.') + [i] for key, value in template_ids.items()
         for i in range(value)])
    seed_ids = [('.'.join(seed_id[0:-1]), seed_id[-1]) for seed_id in seed_ids]
    Logger.info(f"Prepping for {len(seed_ids)} channels that share seed-ids "
                f"between templates and stream")
    Logger.debug(f"Shared seed-ids: {seed_ids}")

    for channel_number, seed_id in enumerate(template_ids.keys()):
        stream_data = np.zeros(stream_length, dtype=np.float32)
        stream_channel = stream.select(id=seed_id)
        if len(stream_channel) > 1:
            msg = f"Multiple channels in continuous data for {seed_id}"
            Logger.error(msg)
            raise NotImplementedError(msg)
        stream_channel = stream_channel[0]
        if stream_channel.stats.npts == stream_length:
            stream_data = stream_channel.data
        else:
            Logger.info('Data for {0} is not as long as needed, '
                        'padding'.format(stream_channel.id))
            if force_stream_epoch:
                start_pad = int(samp_rate * (
                        stream_channel.stats.starttime - stream_start))
                end_pad = stream_length - (
                        start_pad + stream_channel.stats.npts)
                # In some cases there will be one sample missing when sampling
                # time-stamps are not set consistently between channels, this
                # results in start_pad and end_pad being len==0
                if start_pad == 0 and end_pad == 0:
                    Logger.debug("Start and end pad are both zero, padding "
                                 "at one end")
                    if (stream_channel.stats.starttime - stream_start) > (
                       stream_end - stream_channel.stats.endtime):
                        start_pad = int(
                            stream_length - stream_channel.stats.npts)
                    else:
                        end_pad = int(
                            stream_length - stream_channel.stats.npts)
                stream_channel.stats.starttime -= (start_pad / samp_rate)
            else:
                start_pad = 0
                end_pad = stream_length - stream_channel.stats.npts
            if end_pad == 0:
                stream_data[start_pad:] = stream_channel.data
            else:
                stream_data[start_pad:-end_pad] = stream_channel.data
        header = stream_channel.stats.copy()
        header.npts = stream_length
        out_stream += Trace(data=stream_data, header=header)

    # Initialize nan template for speed.
    nan_channel = np.full(template_length, np.nan, dtype=np.float32)
    nan_channel = np.require(nan_channel, requirements=['C_CONTIGUOUS'])
    nan_template = Stream()
    for _seed_id in seed_ids:
        net, sta, loc, chan = _seed_id[0].split('.')
        nan_template += Trace(header=Stats({
            'network': net, 'station': sta, 'location': loc,
            'channel': chan, 'starttime': UTCDateTime(ns=0),
            'npts': template_length, 'sampling_rate': samp_rate}))

    # Remove templates with no matching channels
    filt = np.ones(len(template_names)).astype(bool)
    for i, template in enumerate(templates):
        trace_ids = {tr.id for tr in template}
        if len(trace_ids.intersection(stream_ids)) == 0:
            filt[i] = 0

    _out = dict(zip(
        [_tn for _tn, _filt in zip(template_names, filt) if _filt],
        [_t for _t, _filt in zip(templates, filt) if _filt]))
    flt_templates = list(_out.values())

    if len(_out) != len(templates):
        Logger.debug("Some templates not used due to no matching channels")

    # Ensure that the templates' earliest traces are kept, even if there is no
    # continuous data for them. If this happens, we need to add a NaN-stream to
    # the continuous data to avoid inconsistent detection times.
    n_template_traces = np.array([len(temp) for temp in flt_templates])
    n_stream_traces = sum([n+1 for s, n in seed_ids])
    # These checks are not necessary if all templates will get NaN-traces,
    # because the NaN-traces will save the right starttime for the template.
    nan_stream_ids = list()
    if any(n_template_traces > n_stream_traces):
        earliest_templ_trace_ids = set(
            [template.sort(['starttime'])[0].id for template in flt_templates])
        for earliest_templ_trace_id in earliest_templ_trace_ids:
            if earliest_templ_trace_id not in template_ids:
                nan_stream_ids.append(earliest_templ_trace_id)
                net, sta, loc, chan = earliest_templ_trace_id.split('.')
                nan_template += Trace(header=Stats({
                    'network': net, 'station': sta, 'location': loc,
                    'channel': chan, 'starttime': UTCDateTime(ns=0),
                    'sampling_rate': samp_rate}))
                stream_nan_data = np.full(
                    stream_length, np.nan, dtype=np.float32)
                out_stream += Trace(
                    data=np.ma.masked_array(stream_nan_data, stream_nan_data),
                    header=Stats({
                        'network': net, 'station': sta, 'location': loc,
                        'channel': chan, 'starttime': stream_start,
                        'npts': stream_length, 'sampling_rate': samp_rate}))
                seed_ids.append((earliest_templ_trace_id, 0))

    incomplete_templates = {
        template_name for template_name, template in _out.items() if
        sorted([tr.id for tr in template]) != [tr.id for tr in nan_template]}

    # Fill out the templates with nan channels
    for template_name in incomplete_templates:
        template = _out[template_name]
        template_starttime = min(tr.stats.starttime for tr in template)
        out_template = _quick_copy_stream(nan_template, deepcopy_data=False)

        # Select traces very quickly: assume that trace order does not change,
        # make dict of trace-ids and list of indices and use indices to select
        stream_trace_id_dict = defaultdict(list)
        for n, tr in enumerate(template.traces):
            stream_trace_id_dict[tr.id].append(n)

        for channel_number, _seed_id in enumerate(seed_ids):
            seed_id, channel_index = _seed_id
            # Select all traces with same seed_id, based on indices for
            # corresponding traces stored in stream_trace_id_dict
            # Much quicker than: template_channel = template.select(id=seed_id)
            template_channel = Stream([
                template.traces[idx] for idx in stream_trace_id_dict[seed_id]])
            if len(template_channel) <= channel_index:
                # out_template[channel_number].data = nan_channel  # quicker:
                out_template.traces[channel_number].__dict__[
                    'data'] = np.copy(nan_channel)
                out_template.traces[channel_number].stats.__dict__[
                    'npts'] = template_length
                out_template.traces[channel_number].stats.__dict__[
                    'starttime'] = template_starttime
                out_template.traces[channel_number].stats.__dict__[
                    'endtime'] = UTCDateTime(ns=int(
                        round(template_starttime.ns
                              + (template_length / samp_rate) * 1e9)))
            else:
                out_template.traces[channel_number] = template_channel.traces[
                    channel_index]

        # If a template-trace matches a NaN-trace in the stream , then set
        # template-trace to NaN so that this trace does not appear in channel-
        # list of detections.
        if len(nan_stream_ids) > 0:
            for tr in out_template:
                if tr.id in nan_stream_ids:
                    tr.data = nan_channel
        _out.update({template_name: out_template})

    out_templates = list(_out.values())
    out_template_names = list(_out.keys())

    if named:
        return out_stream, out_templates, out_template_names
    return out_stream, out_templates


def shortproc(st, lowcut, highcut, filt_order, samp_rate, parallel=False,
              num_cores=False, starttime=None, endtime=None,
              seisan_chan_names=False, fill_gaps=True, ignore_length=False,
              ignore_bad_data=False, fft_threads=1):
    """
    Deprecated
    """
    Logger.warning("Shortproc is depreciated after 0.4.4 and will "
                   "be removed in a future version. Use multi_process"
                   " instead")
    st = multi_process(
        st=st, lowcut=lowcut, highcut=highcut, filt_order=filt_order,
        samp_rate=samp_rate, parallel=parallel, num_cores=num_cores,
        starttime=starttime, endtime=endtime, daylong=False,
        seisan_chan_names=seisan_chan_names, fill_gaps=fill_gaps,
        ignore_length=ignore_length, ignore_bad_data=ignore_bad_data)
    return st


def dayproc(st, lowcut, highcut, filt_order, samp_rate, starttime,
            parallel=True, num_cores=False, ignore_length=False,
            seisan_chan_names=False, fill_gaps=True, ignore_bad_data=False,
            fft_threads=1):
    """
    Deprecated
    """
    Logger.warning("dayproc is depreciated after 0.4.4 and will be "
                   "removed in a future version. Use multi_process instead")
    st = multi_process(
        st=st, lowcut=lowcut, highcut=highcut, filt_order=filt_order,
        samp_rate=samp_rate, parallel=parallel, num_cores=num_cores,
        starttime=starttime, endtime=None, daylong=True,
        seisan_chan_names=seisan_chan_names, fill_gaps=fill_gaps,
        ignore_length=ignore_length, ignore_bad_data=ignore_bad_data)
    return st


def process(tr, lowcut, highcut, filt_order, samp_rate,
            starttime=False, clip=False, length=86400,
            seisan_chan_names=False, ignore_length=False, fill_gaps=True,
            ignore_bad_data=False, fft_threads=1):
    """
    Deprecated
    """
    Logger.warning("process is depreciated after 0.4.4 and will be removed "
                   "in a future version. Use multi_process instead")
    if length == 86400:
        daylong = True
    else:
        daylong = False

    endtime = None
    if clip:
        if not starttime:
            starttime = tr.stats.starttime
        elif not isinstance(starttime, UTCDateTime):
            starttime = UTCDateTime(starttime)
        endtime = starttime + length
    st = multi_process(
        st=tr, lowcut=lowcut, highcut=highcut, filt_order=filt_order,
        samp_rate=samp_rate, parallel=False, num_cores=1,
        starttime=starttime, endtime=endtime, daylong=daylong,
        seisan_chan_names=seisan_chan_names, fill_gaps=fill_gaps,
        ignore_length=ignore_length, ignore_bad_data=ignore_bad_data)
    return st


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Functions for network matched-filter detection of seismic data.

Designed to cross-correlate templates generated by template_gen function
with data and output the detections.

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import os
import pickle
import logging
import numpy as np

from collections import defaultdict
from typing import List, Set
from timeit import default_timer

from concurrent.futures import ThreadPoolExecutor

from obspy import Stream, UTCDateTime

from eqcorrscan.core.match_filter.template import (
    Template, group_templates_by_seedid)
from eqcorrscan.core.match_filter.detection import Detection
from eqcorrscan.core.match_filter.party import Party
from eqcorrscan.core.match_filter.family import Family
from eqcorrscan.core.match_filter.helpers import _spike_test, _mad
from eqcorrscan.core.match_filter.matched_filter import MatchFilterError

from eqcorrscan.utils.correlate import (
    get_stream_xcorr, _stabalised_fmf, fftw_multi_normxcorr,
    _zero_invalid_correlation_sums, _set_inner_outer_threading)
from eqcorrscan.utils.pre_processing import (
    _check_daylong, _group_process)
from eqcorrscan.utils.findpeaks import multi_find_peaks
from eqcorrscan.utils.plotting import _match_filter_plot

Logger = logging.getLogger(__name__)


def _wildcard_fill(
    net: str, sta: str, loc: str, chan: str
) -> [str, str, str, str]:
    """
    Convert none to wildcards. Cope with seisan channel naming.

    .. rubric:: Example

    >>> _wildcard_fill(None, None, None, None)
    ('*', '*', '*', '*')
    >>> _wildcard_fill("NZ", "FOZ", "10", "HZ")
    ('NZ', 'FOZ', '10', 'H?Z')
    """
    if net in [None, '']:
        net = "*"
    if sta in [None, '']:
        sta = "*"
    if loc in [None, '']:
        loc = "*"
    if chan in [None, '']:
        chan = "*"
    # Cope with old seisan chans
    if len(chan) == 2:
        chan = f"{chan[0]}?{chan[-1]}"
    return net, sta, loc, chan


def _download_st(
    starttime: UTCDateTime,
    endtime: UTCDateTime,
    buff: float,
    min_gap: float,
    template_channel_ids: List[tuple],
    client,
    retries: int
) -> Stream:
    """
    Helper to download a stream from a client for a given start and end time.

    Applies `buff` to extend download to (heopfully) ensure all data are
    provided. Retries download up to `retries` times, and discards data
    with large gaps.

    :param starttime: Start time to download data from
    :param endtime: End time to download data to
    :param buff:
        Length to pad downloaded data by - some clients do not provide all
        data requested.
    :param min_gap: See core.match_filter.tribe.client_detect
    :param template_channel_ids:
    :param client:
        Client-like object with at least a .get_waveforms_bulk method.
    :param retries: See core.match_filter.tribe.client_detect

    :return: Stream as downloaded.
    """
    from obspy.clients.fdsn.header import FDSNException

    bulk_info = []
    for chan_id in template_channel_ids:
        bulk_info.append((
            chan_id[0], chan_id[1], chan_id[2], chan_id[3],
            starttime - buff, endtime + buff))

    for retry_attempt in range(retries):
        try:
            Logger.info(f"Downloading data between {starttime} and "
                        f"{endtime}")
            st = client.get_waveforms_bulk(bulk_info)
            Logger.info(
                "Downloaded data for {0} traces".format(len(st)))
            break
        except FDSNException as e:
            if "Split the request in smaller" in " ".join(e.args):
                Logger.warning(
                    "Datacentre does not support large requests: "
                    "splitting request into smaller chunks")
                st = Stream()
                for _bulk in bulk_info:
                    try:
                        st += client.get_waveforms_bulk([_bulk])
                    except Exception as e:
                        Logger.error("No data for {0}".format(_bulk))
                        Logger.error(e)
                        continue
                Logger.info("Downloaded data for {0} traces".format(
                    len(st)))
                break
        except Exception as e:
            Logger.error(e)
            continue
    else:
        raise MatchFilterError(
            "Could not download data after {0} attempts".format(
                retries))
    # Get gaps and remove traces as necessary
    if min_gap:
        gaps = st.get_gaps(min_gap=min_gap)
        if len(gaps) > 0:
            Logger.warning("Large gaps in downloaded data")
            st.merge()
            gappy_channels = list(
                set([(gap[0], gap[1], gap[2], gap[3])
                     for gap in gaps]))
            _st = Stream()
            for tr in st:
                tr_stats = (tr.stats.network, tr.stats.station,
                            tr.stats.location, tr.stats.channel)
                if tr_stats in gappy_channels:
                    Logger.warning(
                        "Removing gappy channel: {0}".format(tr))
                else:
                    _st += tr
            st = _st
            st.split()
    # Merge traces after gap checking
    st = st.merge()
    st.trim(starttime=starttime, endtime=endtime)

    st_ids = [tr.id for tr in st]
    # Remove traces that do not meet zero criteria
    st.traces = [tr for tr in st if _check_daylong(tr.data)]
    if len(st) < len(st_ids):
        lost_ids = " ".join([tr_id for tr_id in st_ids
                             if tr_id not in [tr.id for tr in st]])
        Logger.warning(
            f"Removed data for {lost_ids} due to more zero datapoints "
            f"than non-zero.")

    st_ids = [tr.id for tr in st]
    # Remove short traces
    st.traces = [
        tr for tr in st
        if tr.stats.endtime - tr.stats.starttime > 0.8 * (endtime - starttime)]
    if len(st) < len(st_ids):
        lost_ids = " ".join([tr_id for tr_id in st_ids
                             if tr_id not in [tr.id for tr in st]])
        Logger.warning(
            f"Removed data for {lost_ids} due to less than 80% of the "
            f"required length.")

    return st


def _pre_process(
    st: Stream,
    template_ids: set,
    pre_processed: bool,
    filt_order: int,
    highcut: float,
    lowcut: float,
    samp_rate: float,
    process_length: float,
    parallel: bool,
    cores: int,
    daylong: bool,
    ignore_length: bool,
    ignore_bad_data: bool,
    overlap: float, **kwargs
) -> Stream:
    """
    Basic matched-filter processing flow. Data are processing in-place.

    :param st: Stream to process
    :param template_ids:
        Iterable of seed ids in the template set. Only channels matching these
        seed ids will be retained.
    :param pre_processed: See core.match_filter.tribe.detect
    :param filt_order: See utils.pre_processing.multi_process
    :param highcut: See utils.pre_processing.multi_process
    :param lowcut: See utils.pre_processing.multi_process
    :param samp_rate: See utils.pre_processing.multi_process
    :param process_length: See utils.pre_processing.multi_process
    :param parallel: See utils.pre_processing.multi_process
    :param cores: See utils.pre_processing.multi_process
    :param daylong: See utils.pre_processing.multi_process
    :param ignore_length: See utils.pre_processing.multi_process
    :param overlap: See core.match_filter.tribe.detect
    :param ignore_bad_data: See utils.pre_processing.multi_process
    :param template_ids:

    :return: Processed stream
    """
    # Retain only channels that have matches in templates
    Logger.info(template_ids)
    st = Stream([tr for tr in st if tr.id in template_ids])
    Logger.info(f"Processing {(len(st))} channels")
    if len(st) == 0:
        raise IndexError(
            "No matching channels between stream and templates")
    tic = default_timer()
    _spike_test(st)
    toc = default_timer()
    Logger.info(f"Checking for spikes took {toc - tic:.4f} s")
    if not pre_processed:
        st_chunks = _group_process(
            filt_order=filt_order,
            highcut=highcut,
            lowcut=lowcut,
            samp_rate=samp_rate,
            process_length=process_length,
            parallel=parallel,
            cores=cores,
            stream=st,
            daylong=daylong,
            ignore_length=ignore_length,
            overlap=overlap,
            ignore_bad_data=ignore_bad_data)
    else:
        st_chunks = [st]
    Logger.info(f"Stream has been split into {len(st_chunks)} chunks")
    return st_chunks


def _group(
    sids: Set[str],
    templates: List[Template],
    group_size: int,
    groups: List[List[str]] = None
) -> List[List[Template]]:
    """
    Group templates either by seed id, or using pre-computed groups

    :param sids: Seed IDs available in stream
    :param templates: Templates to group
    :param group_size: Maximum group size
    :param groups: [Optional] List of List of template names in groups
    :return: Groups of templates.
    """
    Logger.info(f"Grouping for {sids}")
    if groups:
        Logger.info("Using pre-computed groups")
        t_dict = {t.name: t for t in templates}
        template_groups = []
        for grp in groups:
            template_group = [
                t_dict.get(t_name) for t_name in grp
                if t_name in t_dict.keys()]
            if len(template_group):
                template_groups.append(template_group)
        return template_groups
    template_groups = group_templates_by_seedid(
        templates=templates,
        st_seed_ids=sids,
        group_size=group_size)
    if len(template_groups) == 1 and len(template_groups[0]) == 0:
        Logger.error("No matching ids between stream and templates")
        raise IndexError("No matching ids between stream and templates")
    return template_groups


def _corr_and_peaks(
    templates: List[Template],
    template_names: List[str],
    stream: Stream,
    xcorr_func: str,
    concurrency: str,
    cores: int,
    i: int,
    export_cccsums: bool,
    parallel: bool,
    peak_cores: int,
    threshold: float,
    threshold_type: str,
    trig_int: float,
    sampling_rate: float,
    full_peaks: bool,
    plot: bool,
    plotdir: str,
    plot_format: str,
    prepped: bool = False,
    **kwargs
):
    """
    Compute cross-correlation between templates and a stream. Returns peaks in
     correlation function.

    :param templates: Templates to correlate
    :param template_names: Names of templates (ordered as templates)
    :param stream: Stream to correlate templates with
    :param xcorr_func: Cross-correlation function to use
    :param concurrency: Concurrency of cross-correlation function
    :param cores: Cores (threads) to use for cross-correlation
    :param i: Group-id (internal book-keeping)
    :param export_cccsums: Whether to export the raw cross-correlation sums
    :param parallel: Whether to compute peaks in parallel
    :param peak_cores: Number of cores (threads) to use for peak finding
    :param threshold: Threshold value (user-defined)
    :param threshold_type: Threshold type (e.g. MAD, ...)
    :param trig_int: Trigger interval in seconds
    :param sampling_rate: Sampling rate of data
    :param full_peaks: Whether to compute full peaks, or fast peaks.
    :param plot: Whether to plot correlation sums and peaks or not
    :param plotdir: Where to save plots if made
    :param plot_format: What format (extension) to use for plots.
    :param prepped:
        Whether data have already been prepared for correlation or not.
        If prepped, inputs change for a specific xcorr-function, see code.

    :return: Peaks, thresholds, number of channels, channels for each template
    """
    # Special cases for fmf and fftw to minimize reshaping time.
    Logger.info(
        f"Starting correlation run for template group {i}")
    tic = default_timer()
    if prepped and xcorr_func == "fmf":
        assert isinstance(templates, np.ndarray)
        assert isinstance(stream, np.ndarray)
        # These need to be passed from queues.
        pads = kwargs.get('pads')
        weights = kwargs.get('weights')
        chans = kwargs.get("chans")
        no_chans = kwargs.get("no_chans")
        # We do not care about removing the gain from our data, we copied it.
        multipliers = np.ones((len(stream), 1))
        step = 1  # We only implement single-step correlations
        if concurrency in ("multithread", "multiprocess"):
            arch = "cpu"
        else:
            arch = "gpu"
        cccsums = _stabalised_fmf(
            template_arr=templates, data_arr=stream, weights=weights,
            pads=pads, arch=arch, multipliers=multipliers, step=step)
    elif prepped and xcorr_func in ("fftw", None):
        assert isinstance(templates, dict)
        assert isinstance(stream, dict)
        pads = kwargs.pop('pads')
        seed_ids = kwargs.pop("seed_ids")
        num_cores_inner, num_cores_outer = _set_inner_outer_threading(
            kwargs.get('cores', None), kwargs.get("cores_outer", None),
            len(stream))

        cccsums, tr_chans = fftw_multi_normxcorr(
            template_array=templates, stream_array=stream,
            pad_array=pads, seed_ids=seed_ids, cores_inner=num_cores_inner,
            cores_outer=num_cores_outer, stack=True, **kwargs)
        n_templates = len(cccsums)
        # Post processing
        no_chans = np.sum(np.array(tr_chans).astype(int), axis=0)
        chans = [[] for _i in range(n_templates)]
        for seed_id, tr_chan in zip(seed_ids, tr_chans):
            for chan, state in zip(chans, tr_chan):
                if state:
                    chan.append(seed_id)
        cccsums = _zero_invalid_correlation_sums(cccsums, pads, chans)
        chans = [[(seed_id.split('.')[1], seed_id.split('.')[-1].split('_')[0])
                  for seed_id in _chans] for _chans in chans]
    else:
        # The default just uses stream xcorr funcs.
        multichannel_normxcorr = get_stream_xcorr(xcorr_func, concurrency)
        cccsums, no_chans, chans = multichannel_normxcorr(
            templates=templates, stream=stream, cores=cores, **kwargs
        )
    if len(cccsums[0]) == 0:
        raise MatchFilterError(
            f"Correlation has not run for group {i}, "
            f"zero length cccsum")
    toc = default_timer()
    Logger.info(
        f"Correlations for group {i} of {len(template_names)} "
        f"templates took {toc - tic:.4f} s")
    Logger.debug(
        f"The shape of the returned cccsums in group {i} "
        f"is: {cccsums.shape}")
    Logger.debug(
        f'This is from {len(templates)} templates correlated with '
        f'{len(stream)} channels of data in group {i}')

    # Handle saving correlation stats
    if export_cccsums:
        for i, cccsum in enumerate(cccsums):
            fname = (
                f"{template_names[i]}-{stream[0].stats.starttime}-"
                f"{stream[0].stats.endtime}_cccsum.npy")
            np.save(file=fname, arr=cccsum)
            Logger.info(
                f"Saved correlation statistic to {fname}")

    # Zero mean check
    if np.any(np.abs(cccsums.mean(axis=-1)) > 0.05):
        Logger.warning(
            'Mean of correlations is non-zero!  Check this!')
    if parallel:
        Logger.info(f"Finding peaks using {peak_cores} threads")
    else:
        Logger.info("Finding peaks in serial")
    # This is in the main process because transferring
    #  lots of large correlation sums in queues is very slow
    all_peaks, thresholds = _threshold(
        cccsums=cccsums, no_chans=no_chans,
        template_names=template_names, threshold=threshold,
        threshold_type=threshold_type,
        trig_int=int(trig_int * sampling_rate),
        parallel=parallel, full_peaks=full_peaks,
        peak_cores=peak_cores, plot=plot, stream=stream,
        plotdir=plotdir, plot_format=plot_format)
    return all_peaks, thresholds, no_chans, chans


def _threshold(
    cccsums: np.ndarray,
    no_chans: list,
    template_names: list,
    threshold: float,
    threshold_type: str,
    trig_int: int,  # converted to samples before getting to this func.
    parallel: bool,
    full_peaks: bool,
    peak_cores: int,
    plot: bool,
    stream: Stream,
    plotdir: str,
    plot_format: str,
):
    """
    Find peaks within correlation functions for given thresholds.

    :param cccsums: Numpy array of correlations [templates x samples]
    :param no_chans:
        Number of channels for each correlation (ordered as cccsums)
    :param template_names:
        Template names for each correlation (ordered as cccsums)
    :param threshold: Input threshold value
    :param threshold_type: Input threshold type (e.g. MAD, ...)
    :param trig_int: Trigger interval in SAMPLES.
    :param parallel: Whether to compute peaks in parallel
    :param full_peaks: Whether to compute full peaks or not
    :param peak_cores: Number of cores (threads) to use for peak finding.
    :param plot: Whether to plot the peak finding
    :param stream: Stream for plotting (not needed otherwise)
    :param plotdir: Directory to write plots to
    :param plot_format: Format to save plots in

    :return: (all peaks, used thresholds)
    """
    Logger.debug(f"Got cccsums shaped {cccsums.shape}")
    Logger.debug(f"From {len(template_names)} templates")

    tic = default_timer()
    if str(threshold_type) == str("absolute"):
        thresholds = [threshold for _ in range(len(cccsums))]
    elif str(threshold_type) == str('MAD'):
        median_cores = min([peak_cores, len(cccsums)])
        if cccsums.size < 2e7:  # parallelism not worth it
            median_cores = 1
        with ThreadPoolExecutor(max_workers=median_cores) as executor:
            # Because numpy releases GIL threading can use
            # multiple cores
            medians = executor.map(_mad, cccsums,
                                   chunksize=len(cccsums) // median_cores)
        thresholds = [threshold * median for median in medians]
    else:
        thresholds = [threshold * no_chans[i]
                      for i in range(len(cccsums))]
    toc = default_timer()
    Logger.info(f"Computing thresholds took {toc - tic: .4f} s")
    outtic = default_timer()
    all_peaks = multi_find_peaks(
        arr=cccsums, thresh=thresholds, parallel=parallel,
        trig_int=trig_int, full_peaks=full_peaks, cores=peak_cores)
    outtoc = default_timer()
    Logger.info(f"Finding peaks for group took {outtoc - outtic:.4f}s")

    # Plotting
    if plot and stream:
        for i, cccsum in enumerate(cccsums):
            _match_filter_plot(
                stream=stream, cccsum=cccsum,
                template_names=template_names,
                rawthresh=thresholds[i], plotdir=plotdir,
                plot_format=plot_format, i=i)
        else:
            Logger.error("Plotting enabled but not stream found to plot")

    return all_peaks, thresholds


def _detect(
    template_names: List[str],
    all_peaks: np.ndarray,
    starttime: UTCDateTime,
    delta: float,
    no_chans: List[int],
    chans: List[List[str]],
    thresholds: List[float]
) -> List[Detection]:
    """
    Convert peaks to Detection objects

    :param template_names: Lis of template names
    :param all_peaks: Array of peaks orders as template_names
    :param starttime: Starttime for peak index relative time
    :param delta: Sample interval to convert peaks from samples to time
    :param no_chans: Number of channels used (ordered as template_names)
    :param chans: Channels used (ordered as template_names)
    :param thresholds: Thresholds used (ordered as template_names)

    :return: List of detections.
    """
    tic = default_timer()
    detections = []
    for i, template_name in enumerate(template_names):
        if not all_peaks[i]:
            Logger.debug(f"Found 0 peaks for template {template_name}")
            continue
        Logger.debug(f"Found {len(all_peaks[i])} detections "
                     f"for template {template_name}")
        for peak in all_peaks[i]:
            detecttime = starttime + (peak[1] * delta)
            if peak[0] > no_chans[i]:
                Logger.error(f"Correlation sum {peak[0]} exceeds "
                             f"bounds ({no_chans[i]}")
            detection = Detection(
                template_name=template_name, detect_time=detecttime,
                no_chans=no_chans[i], detect_val=peak[0],
                threshold=thresholds[i], typeofdet='corr',
                chans=chans[i],
                threshold_type=None,
                # Update threshold_type and threshold outside of this func.
                threshold_input=None)
            detections.append(detection)
    toc = default_timer()
    Logger.info(f"Forming detections took {toc - tic:.4f} s")
    return detections


def _load_template(t_file: str) -> Template:
    """ Load a pickled template from a file """
    try:
        with open(t_file, "rb") as f:
            t = pickle.load(f)
    except Exception as e:
        Logger.warning(f"Could not read template from {t_file} due to {e}")
        return None
    assert isinstance(t, Template), "Loaded object is not a Template, aborting"
    return t


def _read_template_db(template_file_dict: dict) -> List[Template]:
    """
    Read templates from files on disk.

    :param template_file_dict: Template file names keyed by template name

    :returns: list of templates
    """
    with ThreadPoolExecutor() as executor:
        templates = executor.map(_load_template, template_file_dict.values())
    templates = [t for t in templates if t]
    Logger.info(f"Deserialized {len(templates)} templates")
    if len(templates) < len(template_file_dict):
        Logger.warning(f"Expected {len(template_file_dict)} templates, "
                       f"but found {len(templates)}")
    return templates


def _make_party(
    detections: List[Detection],
    threshold: float,
    threshold_type: str,
    templates: List[Template],
    chunk_start: UTCDateTime,
    chunk_id: int,
    save_progress: bool
) -> str:
    """
    Construct a Party from Detections.

    :param detections: List of detections
    :param threshold: Input threshold
    :param threshold_type: Input threshold type
    :param templates: Templates used in detections
    :param chunk_start: Starttime of party epoch
    :param chunk_id: Internal index for party epoch
    :param save_progress: Whether to save progress or not

    :return: The filename the party has been pickled to.
    """
    chunk_dir = os.path.join(
        ".parties", "{chunk_start.year}", "{chunk_start.julday:03d}")
    chunk_file_str = os.path.join(
        chunk_dir, "chunk_party_{chunk_start_str}_{chunk_id}_{pid}.pkl")
    # Process ID included in chunk file to avoid multiple processes writing
    # and reading and removing the same files.

    # Get the results out of the end!
    Logger.info(f"Made {len(detections)} detections")

    # post - add in threshold, threshold_type to all detections
    Logger.info("Adding threshold to detections")
    for detection in detections:
        detection.threshold_input = threshold
        detection.threshold_type = threshold_type

    # Select detections very quickly: detection order does not
    # change, make dict of keys: template-names and values:
    # list of indices and use indices to select
    Logger.info("Making dict of detections")
    detection_idx_dict = defaultdict(list)
    for n, detection in enumerate(detections):
        detection_idx_dict[detection.template_name].append(n)

    # Convert to Families and build party.
    Logger.info("Converting to party and making events")
    chunk_party = Party()

    # Make a dictionary of templates keyed by name - we could be passed a dict
    # of pickled templates
    if not isinstance(templates, dict):
        templates = {t.name: t for t in templates}

    for t_name, template in templates.items():
        family_detections = [
            detections[idx]
            for idx in detection_idx_dict[t_name]]
        # Make party sparse - only write out families with detections
        if len(family_detections):
            if not isinstance(template, Template):
                # Try and read this from disk
                with open(template, "rb") as f:
                    template = pickle.load(f)
            for d in family_detections:
                d._calculate_event(template=template)
            family = Family(
                template=template, detections=family_detections)
            chunk_party += family

    Logger.info("Pickling party")
    if not os.path.isdir(chunk_dir.format(chunk_start=chunk_start)):
        os.makedirs(chunk_dir.format(chunk_start=chunk_start))

    chunk_file = chunk_file_str.format(
        chunk_start_str=chunk_start.strftime("%Y-%m-%dT%H-%M-%S"),
        chunk_start=chunk_start,
        chunk_id=chunk_id, pid=os.getpid())
    with open(chunk_file, "wb") as _f:
        pickle.dump(chunk_party, _f)
    Logger.info("Completed party processing")

    if save_progress:
        Logger.info(f"Written chunk to {chunk_file}")
    return chunk_file


if __name__ == "__main__":
    import doctest

    doctest.testmod()
